from __future__ import annotations

import json
import logging
from pathlib import Path

import numpy as np
import pandas as pd

logger = logging.getLogger("tract7dt.patches")


def _grid_bounds(nx: int, ny: int, ngrid: int, r: int, c: int) -> tuple[int, int, int, int]:
    if not (0 <= r < ngrid and 0 <= c < ngrid):
        raise ValueError("r/c out of range")
    x_step = nx // ngrid
    y_step = ny // ngrid
    x0 = c * x_step
    x1 = nx if c == ngrid - 1 else (c + 1) * x_step
    y0 = r * y_step
    y1 = ny if r == ngrid - 1 else (r + 1) * y_step
    return x0, x1, y0, y1


def _grid_bounds_in_bbox(x0: int, x1: int, y0: int, y1: int, ngrid: int, r: int, c: int):
    if not (0 <= r < ngrid and 0 <= c < ngrid):
        raise ValueError("r/c out of range")
    width = x1 - x0
    height = y1 - y0
    if width <= 0 or height <= 0:
        raise ValueError("bbox has non-positive size")

    x_step = width // ngrid
    y_step = height // ngrid
    if x_step <= 0 or y_step <= 0:
        raise ValueError("ngrid too fine for bbox size")

    bx0 = x0 + c * x_step
    bx1 = x1 if c == ngrid - 1 else x0 + (c + 1) * x_step
    by0 = y0 + r * y_step
    by1 = y1 if r == ngrid - 1 else y0 + (r + 1) * y_step
    return bx0, bx1, by0, by1


def build_patches(
    *,
    epsf_ngrid: int,
    patch_ngrid: int,
    final_psf_size: int,
    halo_pix_min: int,
    image_shape: tuple[int, int],
    out_dir: Path,
    active_epsf_tags: set[str] | None = None,
) -> Path:
    out_dir.mkdir(parents=True, exist_ok=True)

    HALO_PIX = max(int((final_psf_size - 1) // 2 + 2), int(halo_pix_min))

    ny, nx = image_shape
    patches = []

    for er in range(epsf_ngrid):
        for ec in range(epsf_ngrid):
            epsf_tag = f"r{er:02d}_c{ec:02d}"
            if active_epsf_tags is not None and epsf_tag not in active_epsf_tags:
                continue
            ex0, ex1, ey0, ey1 = _grid_bounds(nx, ny, epsf_ngrid, er, ec)

            for pr in range(patch_ngrid):
                for pc in range(patch_ngrid):
                    x0b, x1b, y0b, y1b = _grid_bounds_in_bbox(ex0, ex1, ey0, ey1, patch_ngrid, pr, pc)

                    x0r = max(0, x0b - HALO_PIX)
                    x1r = min(nx, x1b + HALO_PIX)
                    y0r = max(0, y0b - HALO_PIX)
                    y1r = min(ny, y1b + HALO_PIX)

                    patch_tag = f"r{er:02d}_c{ec:02d}_pr{pr:02d}_pc{pc:02d}"

                    patches.append(
                        dict(
                            epsf_r=int(er),
                            epsf_c=int(ec),
                            patch_r=int(pr),
                            patch_c=int(pc),
                            epsf_tag=epsf_tag,
                            patch_tag=patch_tag,
                            x0_base=int(x0b),
                            x1_base=int(x1b),
                            y0_base=int(y0b),
                            y1_base=int(y1b),
                            x0_roi=int(x0r),
                            x1_roi=int(x1r),
                            y0_roi=int(y0r),
                            y1_roi=int(y1r),
                            halo_pix=int(HALO_PIX),
                        )
                    )

    patch_df = pd.DataFrame(patches)
    patch_csv = out_dir / "patches.csv"
    patch_json = out_dir / "patches.json"
    patch_df.to_csv(patch_csv, index=False)
    patch_json.write_text(json.dumps(patches, indent=2))

    logger.info("Image shape: %s", (ny, nx))
    logger.info("EPSF_NGRID: %d | PATCH_NGRID: %d | HALO_PIX: %d", epsf_ngrid, patch_ngrid, HALO_PIX)
    if active_epsf_tags is not None:
        logger.info("Active ePSF cells: %d/%d", len(active_epsf_tags), epsf_ngrid * epsf_ngrid)
    logger.info("Total patches: %d", len(patches))
    logger.info("Wrote: %s", patch_csv)
    logger.info("Wrote: %s", patch_json)

    return patch_json



from __future__ import annotations

import hashlib
import logging
import os
import re
import shutil
import sys
import tarfile
import time
import urllib.parse
import urllib.request
import zipfile
from dataclasses import dataclass
from pathlib import Path


_LOG = logging.getLogger("tract7dt.sample_data")
_CHUNK_SIZE = 1024 * 1024  # 1 MiB
_MIB = 1024 * 1024
_FITS_SUFFIXES = (".fits", ".fit", ".fits.fz")
_BAND_RE = re.compile(r"(m\d{3})", re.IGNORECASE)

# Fill these once your GitHub release is published.
DEFAULT_SAMPLE_REPO = "seoldh99/tract7dt_sample_data"
DEFAULT_SAMPLE_TAG = "v1.0.0"
DEFAULT_SAMPLE_ASSET = "tract7dt_sample_data.tar.gz"
DEFAULT_SAMPLE_SHA256 = "91c906114458cc3372241cedcca61d9786aaf1391a40c136a717a02e5349ac94"


@dataclass(frozen=True)
class DownloadResult:
    dataset_dir: Path
    downloaded_file: Path
    image_list_file: Path


def github_release_asset_url(repo: str, tag: str, asset: str) -> str:
    repo = repo.strip().strip("/")
    if "/" not in repo:
        raise ValueError("repo must be in '<owner>/<repo>' format")
    if not tag.strip():
        raise ValueError("tag cannot be empty")
    if not asset.strip():
        raise ValueError("asset cannot be empty")
    return f"https://github.com/{repo}/releases/download/{tag}/{asset}"


def default_sample_data_url() -> str | None:
    repo = os.environ.get("TRACT7DT_SAMPLE_REPO", DEFAULT_SAMPLE_REPO).strip()
    tag = os.environ.get("TRACT7DT_SAMPLE_TAG", DEFAULT_SAMPLE_TAG).strip()
    asset = os.environ.get("TRACT7DT_SAMPLE_ASSET", DEFAULT_SAMPLE_ASSET).strip()
    if not (repo and tag and asset):
        return None
    return github_release_asset_url(repo, tag, asset)


def default_sample_data_sha256() -> str | None:
    value = os.environ.get("TRACT7DT_SAMPLE_SHA256", DEFAULT_SAMPLE_SHA256).strip()
    return value or None


def _infer_filename(url: str) -> str:
    parsed = urllib.parse.urlparse(url)
    name = Path(parsed.path).name
    return name or "sample_data.bin"


def _dataset_name_from_filename(filename: str) -> str:
    p = Path(filename)
    name = p.name
    for ext in (".tar.gz", ".tar.bz2", ".tar.xz", ".tgz", ".zip", ".tar", ".gz"):
        if name.endswith(ext):
            return name[: -len(ext)] or "sample-data"
    return p.stem or "sample-data"


def _sha256(path: Path) -> str:
    h = hashlib.sha256()
    with path.open("rb") as f:
        while True:
            chunk = f.read(_CHUNK_SIZE)
            if not chunk:
                break
            h.update(chunk)
    return h.hexdigest()


def _verify_sha256(path: Path, expected: str | None) -> None:
    if not expected:
        return
    expected_norm = expected.strip().lower()
    got = _sha256(path)
    if got != expected_norm:
        raise RuntimeError(
            f"SHA256 mismatch for {path}. expected={expected_norm}, got={got}"
        )


def _download_url(url: str, out_path: Path) -> None:
    out_path.parent.mkdir(parents=True, exist_ok=True)
    tmp = out_path.with_suffix(out_path.suffix + ".part")
    req = urllib.request.Request(url, headers={"User-Agent": "tract7dt/0.2"})

    with urllib.request.urlopen(req) as resp, tmp.open("wb") as f:
        total = resp.headers.get("Content-Length")
        total_bytes = int(total) if total and total.isdigit() else None
        written = 0
        next_log_pct = 10
        next_log_bytes = 50 * _MIB
        logged_completion = False
        use_inline_progress = sys.stderr.isatty()
        last_render = 0.0
        last_pct = -1

        def _progress_bar(pct: int, width: int = 24) -> str:
            clamped = max(0, min(100, pct))
            filled = int((clamped * width) / 100)
            return "#" * filled + "-" * (width - filled)

        while True:
            chunk = resp.read(_CHUNK_SIZE)
            if not chunk:
                break
            f.write(chunk)
            written += len(chunk)
            if total_bytes:
                pct = int((100 * written) / total_bytes)
                if use_inline_progress:
                    now = time.monotonic()
                    # Refresh in-place roughly 5 times/sec or on completion.
                    if pct != last_pct and (now - last_render >= 0.2 or pct >= 100):
                        msg = (
                            f"\rDownload [{_progress_bar(pct)}] {pct:3d}% "
                            f"({written / _MIB:.1f}/{total_bytes / _MIB:.1f} MiB)"
                        )
                        sys.stderr.write(msg)
                        sys.stderr.flush()
                        last_render = now
                        last_pct = pct
                        if pct >= 100:
                            logged_completion = True
                elif pct >= next_log_pct:
                    _LOG.info(
                        "Download [%s] %3d%% (%.1f/%.1f MiB)",
                        _progress_bar(pct),
                        pct,
                        written / _MIB,
                        total_bytes / _MIB,
                    )
                    next_log_pct += 10
                    if pct >= 100:
                        logged_completion = True
            elif written >= next_log_bytes:
                if use_inline_progress:
                    msg = f"\rDownload [unknown size] {written / _MIB:.1f} MiB downloaded"
                    sys.stderr.write(msg)
                    sys.stderr.flush()
                else:
                    _LOG.info("Download [unknown size] %.1f MiB downloaded", written / _MIB)
                next_log_bytes += 50 * _MIB

        if total_bytes and not logged_completion:
            if use_inline_progress:
                msg = (
                    f"\rDownload [{_progress_bar(100)}] 100% "
                    f"({written / _MIB:.1f}/{total_bytes / _MIB:.1f} MiB)"
                )
                sys.stderr.write(msg)
                sys.stderr.flush()
            else:
                _LOG.info(
                    "Download [%s] 100%% (%.1f/%.1f MiB)",
                    _progress_bar(100),
                    written / _MIB,
                    total_bytes / _MIB,
                )

        if use_inline_progress:
            sys.stderr.write("\n")
            sys.stderr.flush()

    tmp.replace(out_path)


def _archive_roots_zip(path: Path) -> set[str]:
    roots: set[str] = set()
    with zipfile.ZipFile(path, "r") as zf:
        for member in zf.namelist():
            clean = member.strip("/")
            if clean:
                roots.add(clean.split("/", 1)[0])
    return roots


def _archive_roots_tar(path: Path) -> set[str]:
    roots: set[str] = set()
    with tarfile.open(path, "r:*") as tf:
        for member in tf.getmembers():
            clean = member.name.strip("/")
            if clean:
                roots.add(clean.split("/", 1)[0])
    return roots


def _extract_archive(archive: Path, extract_root: Path) -> Path:
    extract_root.mkdir(parents=True, exist_ok=True)
    root = extract_root.resolve()

    def _assert_safe_path(candidate: Path) -> None:
        resolved = candidate.resolve()
        if not str(resolved).startswith(str(root)):
            raise RuntimeError(f"Unsafe archive path detected: {candidate}")

    lower = archive.name.lower()
    roots: set[str]
    if lower.endswith(".zip"):
        roots = _archive_roots_zip(archive)
        with zipfile.ZipFile(archive, "r") as zf:
            for member in zf.namelist():
                _assert_safe_path(extract_root / member)
            zf.extractall(extract_root)
    elif lower.endswith(".tar") or lower.endswith(".tar.gz") or lower.endswith(".tgz"):
        roots = _archive_roots_tar(archive)
        with tarfile.open(archive, "r:*") as tf:
            for member in tf.getmembers():
                _assert_safe_path(extract_root / member.name)
            tf.extractall(extract_root)
    else:
        raise RuntimeError(
            "Sample data archive must be one of .zip/.tar/.tar.gz/.tgz"
        )

    if len(roots) == 1:
        ds_name = next(iter(roots))
        ds = (extract_root / ds_name).resolve()
        if ds.exists() and ds.is_dir():
            return ds

    inferred_name = _dataset_name_from_filename(archive.name)
    guessed = (extract_root / inferred_name).resolve()
    if guessed.exists() and guessed.is_dir():
        return guessed

    return extract_root.resolve()


def _planned_dataset_dir(archive: Path, extract_root: Path) -> Path:
    lower = archive.name.lower()
    roots: set[str] = set()
    if lower.endswith(".zip"):
        roots = _archive_roots_zip(archive)
    elif lower.endswith(".tar") or lower.endswith(".tar.gz") or lower.endswith(".tgz"):
        roots = _archive_roots_tar(archive)

    if len(roots) == 1:
        return (extract_root / next(iter(roots))).resolve()

    return (extract_root / _dataset_name_from_filename(archive.name)).resolve()


def _is_fits_path(path: Path) -> bool:
    lower = path.name.lower()
    return any(lower.endswith(sfx) for sfx in _FITS_SUFFIXES)


def _band_key(band: str) -> int:
    return int(band[1:])


def _extract_band(path: Path) -> str | None:
    m = _BAND_RE.search(path.name)
    return m.group(1).lower() if m else None


def _standardize_image_names(dataset_dir: Path) -> list[Path]:
    images_dir = dataset_dir / "images"
    search_root = images_dir if images_dir.exists() else dataset_dir
    fits_files = [
        p.resolve()
        for p in search_root.rglob("*")
        if p.is_file() and _is_fits_path(p)
    ]
    if not fits_files:
        raise RuntimeError(f"No FITS files found under {search_root}")

    band_to_file: dict[str, Path] = {}
    for p in fits_files:
        band = _extract_band(p)
        if not band:
            continue
        if band in band_to_file:
            raise RuntimeError(f"Duplicate FITS files found for band {band}")
        band_to_file[band] = p

    if not band_to_file:
        raise RuntimeError("Could not infer any band names from FITS filenames")

    ordered_bands = sorted(band_to_file.keys(), key=_band_key)
    out_files: list[Path] = []

    for band in ordered_bands:
        src = band_to_file[band]
        dst = (images_dir / f"7DT_UDS_{band}.fits").resolve()
        images_dir.mkdir(parents=True, exist_ok=True)
        if src != dst:
            if dst.exists():
                dst.unlink()
            shutil.move(str(src), str(dst))
        out_files.append(dst)

    return out_files


def _write_image_list(dataset_dir: Path, image_paths: list[Path]) -> Path:
    out = dataset_dir / "image_list.txt"
    out.write_text("".join(f"{p.resolve()}\n" for p in image_paths))
    return out


def download_sample_data(
    *,
    path_to_download: Path,
    target_dataset_dir: Path | None = None,
    overwrite: bool = False,
) -> DownloadResult:
    url = default_sample_data_url()
    if not url:
        raise RuntimeError(
            "No default sample-data source configured. "
            "Set TRACT7DT_SAMPLE_REPO/TAG/ASSET (or constants in sample_data.py)."
        )
    checksum = default_sample_data_sha256()
    download_root = path_to_download.expanduser().resolve()
    download_root.mkdir(parents=True, exist_ok=True)
    filename = _infer_filename(url)
    archive_path = download_root / filename

    if archive_path.exists():
        if archive_path.is_dir():
            shutil.rmtree(archive_path)
        else:
            archive_path.unlink()
        _LOG.debug("Removed pre-existing archive path: %s", archive_path)

    _LOG.info("Downloading sample data from: %s", url)
    _download_url(url, archive_path)
    _LOG.debug("Downloaded file: %s", archive_path)

    try:
        _verify_sha256(archive_path, checksum)
        if checksum:
            _LOG.debug("SHA256 verified")

        final_target = (
            target_dataset_dir.expanduser().resolve()
            if target_dataset_dir is not None
            else _planned_dataset_dir(archive_path, download_root)
        )
        if target_dataset_dir is not None:
            _LOG.debug("Requested target dataset directory (--dir): %s", final_target)
        if final_target.exists():
            if not overwrite:
                raise RuntimeError(
                    f"Target dataset directory already exists: {final_target}. "
                    "Use --force to overwrite."
                )
            if final_target.is_dir():
                shutil.rmtree(final_target)
            else:
                final_target.unlink()
            _LOG.info("Removed existing target directory due to --force: %s", final_target)

        dataset_dir = _extract_archive(archive_path, download_root)
        _LOG.debug("Extracted archive into staging root: %s", download_root)
        _LOG.debug("Detected extracted dataset directory: %s", dataset_dir)

        if target_dataset_dir is not None:
            target_dir = target_dataset_dir.expanduser().resolve()
            if dataset_dir != target_dir:
                if dataset_dir == download_root:
                    raise RuntimeError(
                        "Cannot rename extracted dataset directory because archive "
                        "does not contain a single top-level folder."
                    )
                if target_dir.exists():
                    raise RuntimeError(f"Target dataset directory already exists: {target_dir}")
                target_dir.parent.mkdir(parents=True, exist_ok=True)
                _LOG.debug(
                    "Moving extracted dataset directory to requested target: %s -> %s",
                    dataset_dir,
                    target_dir,
                )
                shutil.move(str(dataset_dir), str(target_dir))
                dataset_dir = target_dir
            _LOG.debug("Dataset directory ready at requested path: %s", dataset_dir)
        else:
            _LOG.debug("Dataset directory ready at default path: %s", dataset_dir)

        image_paths = _standardize_image_names(dataset_dir)
        image_list_file = _write_image_list(dataset_dir, image_paths)
        _LOG.debug("Generated image list: %s", image_list_file)

        return DownloadResult(
            dataset_dir=dataset_dir,
            downloaded_file=archive_path,
            image_list_file=image_list_file,
        )
    finally:
        if archive_path.exists():
            if archive_path.is_dir():
                shutil.rmtree(archive_path)
            else:
                archive_path.unlink()
            _LOG.debug("Removed temporary archive path: %s", archive_path)



"""
Merge patch-level Tractor results back into a single catalog.
"""

from __future__ import annotations

import argparse
import logging
from pathlib import Path

import numpy as np
import pandas as pd

from .logging_utils import setup_logging

logger = logging.getLogger("tract7dt.merge")


def _pick_key_cols(df: pd.DataFrame) -> list[str]:
    if "ID" in df.columns:
        return ["ID"]
    ra = "RA" if "RA" in df.columns else ("ra" if "ra" in df.columns else None)
    dec = "DEC" if "DEC" in df.columns else ("dec" if "dec" in df.columns else None)
    if ra and dec:
        return [ra, dec]
    raise ValueError("Could not find merge key columns. Need 'ID' or ('RA','DEC').")


def merge_catalogs(
    *,
    input_catalog: Path,
    patch_outdir: Path,
    out_path: Path,
    pattern: str,
    wcs_fits: Path | None,
    exclusion_flags: pd.DataFrame | None = None,
) -> Path:
    if not input_catalog.exists():
        raise SystemExit(f"Missing input catalog: {input_catalog}")
    if not patch_outdir.exists():
        raise SystemExit(f"Missing patch outdir: {patch_outdir}")

    base = pd.read_csv(input_catalog)
    key_cols = _pick_key_cols(base)
    if exclusion_flags is not None and len(exclusion_flags) > 0:
        flag_cols = [c for c in ("excluded_crop", "excluded_saturation") if c in exclusion_flags.columns]
        if flag_cols:
            flags = exclusion_flags.copy()
            for c in flag_cols:
                flags[c] = flags[c].fillna(False).astype(bool)
            flags = flags.groupby(key_cols, as_index=False, dropna=False)[flag_cols].max()
            base = base.merge(flags, how="left", on=key_cols)
            for c in flag_cols:
                base[c] = base[c].fillna(False).astype(bool)
            if "excluded_crop" not in base.columns:
                base["excluded_crop"] = False
            if "excluded_saturation" not in base.columns:
                base["excluded_saturation"] = False
            base["excluded_any"] = base["excluded_crop"] | base["excluded_saturation"]
            base["excluded_reason"] = np.select(
                [
                    base["excluded_crop"] & base["excluded_saturation"],
                    base["excluded_crop"],
                    base["excluded_saturation"],
                ],
                [
                    "crop+saturation",
                    "crop",
                    "saturation",
                ],
                default="",
            )

    csvs = sorted(patch_outdir.glob(f"**/{pattern}"))
    if len(csvs) == 0:
        raise SystemExit(f"No patch result CSVs found under {patch_outdir} with pattern {pattern}")

    rows: list[pd.DataFrame] = []
    for p in csvs:
        df = pd.read_csv(p)
        legacy_force = ("x_fit_white", "y_fit_white")
        fit_cols = [c for c in df.columns if c not in base.columns or c in legacy_force]
        keep_cols = list(dict.fromkeys(key_cols + fit_cols))
        keep_cols = [c for c in keep_cols if c in df.columns]
        rows.append(df[keep_cols])

    fit = pd.concat(rows, ignore_index=True)

    if fit.duplicated(subset=key_cols).any():
        dup = fit.loc[fit.duplicated(subset=key_cols, keep=False), key_cols].head(10)
        raise ValueError(f"Duplicate keys in patch results (showing up to 10):\n{dup}")

    merged = base.merge(fit, how="left", on=key_cols, suffixes=("", "_fitdup"))

    dup_cols = [c for c in merged.columns if c.endswith("_fitdup")]
    if dup_cols:
        merged = merged.drop(columns=dup_cols)

    if wcs_fits:
        want = ("RA_fit" in merged.columns) and ("DEC_fit" in merged.columns)
        have_xy = ("x_pix_white_fit" in merged.columns) and ("y_pix_white_fit" in merged.columns)
        if not want:
            merged["RA_fit"] = np.nan
            merged["DEC_fit"] = np.nan
            want = True
        if want and have_xy:
            try:
                from astropy.io import fits  # type: ignore
                from astropy.wcs import WCS  # type: ignore

                with fits.open(str(wcs_fits), memmap=True) as hdul:
                    wcs = WCS(hdul[0].header)

                xw = pd.to_numeric(merged["x_pix_white_fit"], errors="coerce").to_numpy(dtype=float)
                yw = pd.to_numeric(merged["y_pix_white_fit"], errors="coerce").to_numpy(dtype=float)
                ok = np.isfinite(xw) & np.isfinite(yw)
                ra = merged["RA_fit"].to_numpy(dtype=float, copy=True)
                dec = merged["DEC_fit"].to_numpy(dtype=float, copy=True)
                fill = ok & (~np.isfinite(ra) | ~np.isfinite(dec))
                if np.any(fill):
                    r, d = wcs.all_pix2world(xw[fill], yw[fill], 0)
                    ra[fill] = np.asarray(r, dtype=float)
                    dec[fill] = np.asarray(d, dtype=float)
                    merged["RA_fit"] = ra
                    merged["DEC_fit"] = dec
                    logger.info("Filled RA_fit/DEC_fit using WCS from: %s", wcs_fits)
            except Exception as e:
                logger.warning("failed to compute RA_fit/DEC_fit in merge: %s", str(e))

    out_path.parent.mkdir(parents=True, exist_ok=True)
    merged.to_csv(out_path, index=False)
    logger.info("Wrote: %s", out_path)
    logger.info("rows: %d cols: %d", len(merged), len(merged.columns))
    if "excluded_any" in merged.columns:
        n_crop = int(merged.get("excluded_crop", pd.Series(False, index=merged.index)).astype(bool).sum())
        n_sat = int(merged.get("excluded_saturation", pd.Series(False, index=merged.index)).astype(bool).sum())
        n_any = int(merged["excluded_any"].astype(bool).sum())
        logger.info("exclusions: crop=%d saturation=%d any=%d", n_crop, n_sat, n_any)
    missing = int(merged[key_cols[0]].isna().sum()) if len(key_cols) == 1 else int(np.any(merged[key_cols].isna(), axis=1).sum())
    logger.info("merge-missing (left join rows without fit): %d", missing)
    return out_path


def main(argv: list[str] | None = None) -> int:
    setup_logging()
    ap = argparse.ArgumentParser()
    ap.add_argument("--input-catalog", required=True, help="Input catalog CSV used to build patch inputs")
    ap.add_argument("--patch-outdir", required=True, help="Base output directory containing per-patch results")
    ap.add_argument("--out", required=True, help="Output merged CSV path")
    ap.add_argument("--pattern", default="*_cat_fit.csv", help="Glob pattern for patch result CSVs")
    ap.add_argument("--wcs-fits", default=None, help="Optional FITS path for RA_fit/DEC_fit")
    args = ap.parse_args(argv)

    merge_catalogs(
        input_catalog=Path(args.input_catalog),
        patch_outdir=Path(args.patch_outdir),
        out_path=Path(args.out),
        pattern=str(args.pattern),
        wcs_fits=Path(args.wcs_fits) if args.wcs_fits else None,
    )
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



"""
Orchestrate patch-level Tractor runs via subprocesses.
"""

from __future__ import annotations

import argparse
import concurrent.futures as cf
import logging
import os
import re
import subprocess
import sys
import time
from pathlib import Path

import pandas as pd

from .logging_utils import setup_logging

logger = logging.getLogger("tract7dt.run_patches")


def _subprocess_env(*, threads_per_process: int, package_root: Path) -> dict:
    env = os.environ.copy()
    t = str(int(max(1, threads_per_process)))
    for k in (
        "OMP_NUM_THREADS",
        "OPENBLAS_NUM_THREADS",
        "MKL_NUM_THREADS",
        "NUMEXPR_NUM_THREADS",
        "VECLIB_MAXIMUM_THREADS",
    ):
        env[k] = t
    pkg_path = str(package_root)
    if env.get("PYTHONPATH"):
        env["PYTHONPATH"] = pkg_path + os.pathsep + env["PYTHONPATH"]
    else:
        env["PYTHONPATH"] = pkg_path
    return env


def _patch_tag_from_input_path(p: Path) -> str:
    name = p.name
    m = re.match(r"^(.*)\.pkl\.gz$", name)
    if m:
        return m.group(1)
    return p.stem


def _should_run(p: Path, outdir_base: Path, resume: bool) -> bool:
    if not resume:
        return True
    tag = _patch_tag_from_input_path(p)
    out_csv = outdir_base / tag / f"{tag}_cat_fit.csv"
    return not out_csv.exists()


def _run_one(
    *,
    patch_input: Path,
    runner_module: str,
    epsf_root: Path,
    outdir_base: Path,
    python_exe: str,
    threads_per_process: int,
    resume: bool,
    extra_args: list[str],
    package_root: Path,
) -> dict:
    tag = _patch_tag_from_input_path(patch_input)
    patch_out = outdir_base / tag
    patch_out.mkdir(parents=True, exist_ok=True)

    if resume and (not _should_run(patch_input, outdir_base, resume=True)):
        return {"patch": tag, "status": "skipped", "returncode": 0, "sec": 0.0}

    cmd = [
        python_exe,
        "-m",
        runner_module,
        "--inputs",
        str(patch_input),
        "--epsf-root",
        str(epsf_root),
        "--outdir",
        str(outdir_base),
    ]
    cmd.extend(extra_args)

    t0 = time.time()
    try:
        proc = subprocess.run(
            cmd,
            text=True,
            capture_output=True,
            env=_subprocess_env(threads_per_process=threads_per_process, package_root=package_root),
        )
        dt = time.time() - t0
    except Exception as e:
        dt = time.time() - t0
        runner_log = patch_out / f"{tag}.runner.log"
        runner_log.write_text("CMD: " + " ".join(cmd) + "\n\n" + f"Runner exception: {e}\n")
        return {"patch": tag, "status": "fail", "returncode": -1, "sec": float(dt), "runner_log": str(runner_log)}

    runner_log = patch_out / f"{tag}.runner.log"
    runner_log.write_text(
        "CMD: " + " ".join(cmd) + "\n\n"
        + "--- STDOUT ---\n" + (proc.stdout or "") + "\n\n"
        + "--- STDERR ---\n" + (proc.stderr or "") + "\n"
    )

    return {
        "patch": tag,
        "status": "ok" if proc.returncode == 0 else "fail",
        "returncode": int(proc.returncode),
        "sec": float(dt),
        "runner_log": str(runner_log),
    }


def run_subprocesses(
    *,
    patch_input_dir: Path,
    epsf_root: Path,
    outdir: Path,
    python_exe: str,
    max_workers: int,
    threads_per_process: int,
    resume: bool,
    extra_args: list[str],
) -> Path:
    runner_module = "tract7dt.tractor_patch_runner"
    package_root = Path(__file__).resolve().parents[1]
    outdir.mkdir(parents=True, exist_ok=True)

    if not patch_input_dir.exists():
        raise SystemExit(f"Missing patch input dir: {patch_input_dir}")

    inputs_all = sorted(patch_input_dir.glob("*.pkl.gz"))
    inputs = [p for p in inputs_all if _should_run(p, outdir, resume=bool(resume))]

    logger.info("Found inputs: %d | to run: %d | max_workers=%d", len(inputs_all), len(inputs), int(max_workers))

    results: list[dict] = []
    failures: list[dict] = []
    tstart = time.time()
    use_inline_progress = sys.stderr.isatty()

    def _format_eta(seconds: float) -> str:
        if not (seconds >= 0) or seconds == float("inf"):
            return "--:--"
        s = int(round(seconds))
        m, ss = divmod(s, 60)
        h, mm = divmod(m, 60)
        if h > 0:
            return f"{h:d}:{mm:02d}:{ss:02d}"
        return f"{mm:02d}:{ss:02d}"

    def _progress_bar(done: int, total: int, width: int = 24) -> str:
        if total <= 0:
            return "-" * width
        frac = max(0.0, min(1.0, float(done) / float(total)))
        filled = int(round(frac * width))
        return "#" * filled + "-" * (width - filled)

    with cf.ThreadPoolExecutor(max_workers=int(max_workers)) as ex:
        fut_to_input = {
            ex.submit(
                _run_one,
                patch_input=p,
                runner_module=runner_module,
                epsf_root=epsf_root,
                outdir_base=outdir,
                python_exe=str(python_exe),
                threads_per_process=int(threads_per_process),
                resume=bool(resume),
                extra_args=extra_args,
                package_root=package_root,
            ): p
            for p in inputs
        }
        pending = set(fut_to_input.keys())
        n = len(pending)
        done = 0
        ok = 0
        skipped = 0
        last_non_tty_log = 0.0

        def _emit_progress(force_log: bool = False) -> None:
            nonlocal last_non_tty_log
            elapsed = max(1e-6, time.time() - tstart)
            rate = done / elapsed
            running = len(pending)
            remaining = max(0, n - done)
            eta = float("inf") if rate <= 0 else remaining / rate
            msg = (
                f"Run patches [{_progress_bar(done, n)}] "
                f"{done}/{n} ok={ok} skipped={skipped} fail={len(failures)} "
                f"running={running} rate={rate:.2f}/s eta={_format_eta(eta)}"
            )
            if use_inline_progress:
                sys.stderr.write("\r" + msg)
                sys.stderr.flush()
            else:
                now = time.time()
                if force_log or (now - last_non_tty_log) >= 15.0:
                    logger.info(msg)
                    last_non_tty_log = now

        if n == 0:
            _emit_progress(force_log=True)
        else:
            _emit_progress(force_log=True)
            while pending:
                done_now, pending = cf.wait(pending, timeout=1.0, return_when=cf.FIRST_COMPLETED)
                if not done_now:
                    _emit_progress(force_log=False)
                    continue
                for fut in done_now:
                    res = fut.result()
                    results.append(res)
                    done += 1
                    status = str(res.get("status", ""))
                    if status == "ok":
                        ok += 1
                    elif status == "skipped":
                        skipped += 1
                    else:
                        failures.append(res)
                _emit_progress(force_log=True)

        if use_inline_progress:
            sys.stderr.write("\n")
            sys.stderr.flush()

    df = pd.DataFrame(results)
    summary_csv = outdir / "run_summary.csv"
    df.to_csv(summary_csv, index=False)
    logger.info(
        "DONE. total: %d ok: %s fail: %d",
        len(results),
        int((df["status"] == "ok").sum()) if "status" in df.columns else "?",
        len(failures),
    )
    logger.info("Wrote: %s", summary_csv)
    if failures:
        logger.warning("First few failures:")
        for r in failures[:10]:
            logger.warning(
                "patch=%s rc=%s log=%s",
                r.get("patch"),
                r.get("returncode"),
                r.get("runner_log"),
            )
    return summary_csv


def main(argv: list[str] | None = None) -> int:
    setup_logging()
    ap = argparse.ArgumentParser()
    ap.add_argument("--patch-input-dir", required=True, help="Directory containing patch input .pkl.gz files")
    ap.add_argument("--epsf-root", required=True)
    ap.add_argument("--outdir", required=True)
    ap.add_argument("--max-workers", type=int, default=32)
    ap.add_argument("--threads-per-process", type=int, default=1)
    ap.add_argument("--resume", action="store_true")
    ap.add_argument("--python", default=os.sys.executable)
    ap.add_argument("--no-cutouts", action="store_true")
    ap.add_argument("--no-patch-overview", action="store_true")
    ap.add_argument("--cutout-size", type=int, default=100)
    ap.add_argument("--cutout-max-sources", type=int, default=None)
    ap.add_argument("--cutout-start", type=int, default=0)
    ap.add_argument("--cutout-data-p-lo", type=float, default=1)
    ap.add_argument("--cutout-data-p-hi", type=float, default=99)
    ap.add_argument("--cutout-model-p-lo", type=float, default=1)
    ap.add_argument("--cutout-model-p-hi", type=float, default=99)
    ap.add_argument("--cutout-resid-p-abs", type=float, default=99)
    ap.add_argument("--gal-model", default="dev", choices=["dev", "exp", "sersic", "star"])
    ap.add_argument("--n-opt-iters", type=int, default=200)
    ap.add_argument("--dlnp-stop", type=float, default=1e-6)
    ap.add_argument("--flag-maxiter-margin", type=int, default=5)
    ap.add_argument("--r-ap", type=float, default=5.0)
    ap.add_argument("--eps-flux", type=float, default=1e-4)
    ap.add_argument("--psf-model", default="hybrid", choices=["hybrid", "pixelized", "gaussian_mixture_from_stamp"])
    ap.add_argument("--psf-fallback-model", default="gaussian_mixture", choices=["gaussian_mixture", "ncircular_gaussian", "moffat"])
    ap.add_argument("--min-epsf-nstars-for-use", type=int, default=5)
    ap.add_argument("--ncircular-gaussian-n", type=int, default=3, choices=[1, 2, 3])
    ap.add_argument("--hybrid-psf-n", type=int, default=4)
    ap.add_argument("--gaussian-mixture-n", type=int, default=4)
    ap.add_argument("--sersic-n-init", type=float, default=3.0)
    ap.add_argument("--re-fallback-pix", type=float, default=3.0)
    ap.add_argument("--fallback-default-fwhm-pix", type=float, default=4.0)
    ap.add_argument("--fallback-stamp-radius-pix", type=float, default=25.0)
    ap.add_argument("--moffat-beta", type=float, default=3.5)
    ap.add_argument("--moffat-radius-pix", type=float, default=25.0)
    args = ap.parse_args(argv)

    extra_args: list[str] = [
        "--cutout-size",
        str(int(args.cutout_size)),
        "--cutout-start",
        str(int(args.cutout_start)),
    "--cutout-data-p-lo",
    str(float(args.cutout_data_p_lo)),
    "--cutout-data-p-hi",
    str(float(args.cutout_data_p_hi)),
    "--cutout-model-p-lo",
    str(float(args.cutout_model_p_lo)),
    "--cutout-model-p-hi",
    str(float(args.cutout_model_p_hi)),
    "--cutout-resid-p-abs",
    str(float(args.cutout_resid_p_abs)),
        "--gal-model",
        str(args.gal_model),
        "--n-opt-iters",
        str(int(args.n_opt_iters)),
        "--dlnp-stop",
        str(float(args.dlnp_stop)),
        "--flag-maxiter-margin",
        str(int(args.flag_maxiter_margin)),
    "--r-ap",
    str(float(args.r_ap)),
    "--eps-flux",
    str(float(args.eps_flux)),
    "--psf-model",
    str(args.psf_model),
    "--psf-fallback-model",
    str(args.psf_fallback_model),
    "--min-epsf-nstars-for-use",
    str(int(args.min_epsf_nstars_for_use)),
    "--ncircular-gaussian-n",
    str(int(args.ncircular_gaussian_n)),
    "--hybrid-psf-n",
    str(int(args.hybrid_psf_n)),
    "--gaussian-mixture-n",
    str(int(args.gaussian_mixture_n)),
    "--sersic-n-init",
    str(float(args.sersic_n_init)),
    "--re-fallback-pix",
    str(float(args.re_fallback_pix)),
        "--fallback-default-fwhm-pix",
        str(float(args.fallback_default_fwhm_pix)),
        "--fallback-stamp-radius-pix",
        str(float(args.fallback_stamp_radius_pix)),
        "--moffat-beta",
        str(float(args.moffat_beta)),
        "--moffat-radius-pix",
        str(float(args.moffat_radius_pix)),
    ]
    if args.cutout_max_sources is not None:
        extra_args.extend(["--cutout-max-sources", str(int(args.cutout_max_sources))])
    if args.no_cutouts:
        extra_args.append("--no-cutouts")
    if args.no_patch_overview:
        extra_args.append("--no-patch-overview")
    run_subprocesses(
        patch_input_dir=Path(args.patch_input_dir),
        epsf_root=Path(args.epsf_root),
        outdir=Path(args.outdir),
        python_exe=str(args.python),
        max_workers=int(args.max_workers),
        threads_per_process=int(args.threads_per_process),
        resume=bool(args.resume),
        extra_args=extra_args,
    )
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

from pathlib import Path
from typing import Any

import yaml


DEFAULT_CONFIG: dict[str, Any] = {
    "inputs": {
        "tile_name": "UDS",
        "input_catalog": "/data2/seoldh/Tractor_Pipeline/input_catalog_from_subaru.csv",
        "image_list_file": "image_list.txt",
        "gaiaxp_synphot_csv": "/data2/seoldh/Tractor_Pipeline/7DT/gaiaxp_dr3_synphot_UDS.csv",
    },
    "outputs": {
        "work_dir": "/data2/seoldh/test_tract7dt",
        "epsf_dir": "EPSFs",
        "patches_dir": "patches",
        "patch_inputs_dir": "patch_payloads",
        "tractor_out_dir": "outputs",
        "final_catalog": "output_catalog.csv",
        "cropped_images_dir": "cropped_images",
    },
    "image_scaling": {
        "zp_ref": 25.0,
    },
    "crop": {
        "enabled": True,
        "margin": 500,
        "plot_pre_crop": True,
        "display_downsample": 4,
        "post_crop_display_downsample": 4,
        "overlay_catalog": True,
        "overlay_downsample_full": 4,
        "overlay_downsample_crop": 4,
        "overlay_crop_box": [4000, 5000, 4000, 5000],
        "overlay_crop_enabled": True,
    },
    "checks": {
        "require_wcs_alignment": True,
        "wcs_tolerance": {
            "crval": 1e-6,
            "crpix": 1e-6,
            "cd": 1e-9,
            "cdelt": 1e-9,
        },
        "require_same_shape": True,
    },
    "source_saturation_cut": {
        "enabled": True,
        "radius_pix": 20.0,
        "require_all_bands": False,
    },
    "logging": {
        "ignore_warnings": True,
        "level": "INFO",
        "file": None,
        "format": "%(asctime)s | %(name)s | %(levelname)s | %(message)s",
    },
    "performance": {
        "frame_prep_workers": "auto",
        "white_stack_workers": "auto",
    },
    "epsf": {
        "use_gaiaxp": True,
        "epsf_ngrid": 7,
        "ngrid": 6,
        "skip_empty_epsf_patches": True,
        "thresh_sigma": 10.0,
        "minarea": 25,
        "cutout_size": 61,
        "edge_pad": 10,
        "min_sep": 30.0,
        "max_stars": 30,
        "q_range": [0.95, 1.05],
        "psfstar_mode": "sep+gaia",
        "gaia_mag_min": 10.0,
        "gaia_mag_max": 30.0,
        "gaia_snap_to_sep": True,
        "gaia_forced_k_fwhm": 1.5,
        "gaia_forced_rmin_pix": 3.0,
        "gaia_forced_rmax_pix": 15.0,
        "gaia_snr_min": 20.0,
        "gaia_match_r_pix": 5.0,
        "gaia_seed_sat_r": 5,
        "gaia_seed_sat_div": 1.3,
        "oversamp": 1,
        "maxiters": 30,
        "do_clip": True,
        "recenter_boxsize": 9,
        "final_psf_size": 55,
        "save_star_montage_max": 100,
        "parallel_bands": True,
        "max_workers": "auto",
    },
    "patches": {
        "halo_pix_min": 20,
    },
    "patch_inputs": {
        "save_patch_inputs": True,
        "skip_empty_patch": True,
        "include_wcs_in_payload": False,
        "max_workers": 16,
        "gzip_compresslevel": 1,
        "keep_payloads_in_memory": False,
    },
    "patch_run": {
        "python_exe": "python",
        "resume": False,
        "max_workers": 32,
        "threads_per_process": 1,
        "gal_model": "exp",
        "n_opt_iters": 25,
        "dlnp_stop": 1e-6,
        "r_ap": 5.0,
        "eps_flux": 1e-4,
        "sersic_n_init": 3.0,
        "re_fallback_pix": 3.0,
        "psf_model": "pixelized",
        "psf_fallback_model": "moffat",
        "min_epsf_nstars_for_use": 5,
        "ncircular_gaussian_n": 3,
        "hybrid_psf_n": 9,
        "gaussian_mixture_n": 4,
        "fallback_default_fwhm_pix": 4.0,
        "fallback_stamp_radius_pix": 25.0,
        "cutout_size": 100,
        "cutout_max_sources": None,
        "cutout_start": 0,
        "cutout_data_p_lo": 1,
        "cutout_data_p_hi": 99,
        "cutout_model_p_lo": 1,
        "cutout_model_p_hi": 99,
        "cutout_resid_p_abs": 99,
        "no_cutouts": False,
        "no_patch_overview": False,
        "flag_maxiter_margin": 0,
    },
    "moffat_psf": {
        "fwhm_default_pix": 4.0,
        "beta": 3.5,
        "radius_pix": 25.0,
        "module_path": None,
    },
    "merge": {
        "pattern": "*_cat_fit.csv",
        "wcs_fits": None,
    },
}


def _deep_update(base: dict[str, Any], override: dict[str, Any]) -> dict[str, Any]:
    out = dict(base)
    for k, v in (override or {}).items():
        if isinstance(v, dict) and isinstance(out.get(k), dict):
            out[k] = _deep_update(out[k], v)
        else:
            out[k] = v
    return out


def _resolve_path(value: Any, base_dir: Path) -> Path | None:
    if value is None:
        return None
    p = Path(value).expanduser()
    if not p.is_absolute():
        p = base_dir / p
    return p.resolve()


def _join_path(prefix: str, key: str) -> str:
    return f"{prefix}.{key}" if prefix else key


def _validate_user_config(user_cfg: dict[str, Any], defaults: dict[str, Any], prefix: str = "") -> None:
    if not isinstance(user_cfg, dict):
        raise TypeError("Config root must be a mapping (YAML dict).")
    for k, v in user_cfg.items():
        if k not in defaults:
            raise KeyError(f"Unknown config key: {_join_path(prefix, str(k))}")
        d = defaults[k]
        if isinstance(v, dict) and isinstance(d, dict):
            _validate_user_config(v, d, _join_path(prefix, str(k)))


def _is_int(value: Any) -> bool:
    return isinstance(value, int) and not isinstance(value, bool)


def _validate_config_types(cfg: dict[str, Any], defaults: dict[str, Any], prefix: str = "") -> None:
    type_overrides: dict[str, tuple[type, ...]] = {
        "epsf.max_workers": (int, str),
        "performance.frame_prep_workers": (int, str),
        "performance.white_stack_workers": (int, str),
        "patch_run.cutout_max_sources": (int, type(None)),
        "moffat_psf.module_path": (str, Path, type(None)),
        "merge.wcs_fits": (str, Path, type(None)),
        "logging.file": (str, Path, type(None)),
    }
    for k, d in defaults.items():
        path = _join_path(prefix, str(k))
        if k not in cfg:
            continue
        v = cfg[k]
        if path in type_overrides:
            if not isinstance(v, type_overrides[path]):
                allowed = ", ".join(t.__name__ for t in type_overrides[path])
                raise TypeError(f"Config key '{path}' must be one of: {allowed}")
            continue
        if isinstance(d, dict):
            if not isinstance(v, dict):
                raise TypeError(f"Config key '{path}' must be a mapping (YAML dict).")
            _validate_config_types(v, d, path)
            continue
        if isinstance(d, list):
            if not isinstance(v, list):
                raise TypeError(f"Config key '{path}' must be a list.")
            continue
        if d is None:
            continue
        if isinstance(d, bool):
            if not isinstance(v, bool):
                raise TypeError(f"Config key '{path}' must be a bool.")
            continue
        if isinstance(d, int):
            if not _is_int(v):
                raise TypeError(f"Config key '{path}' must be an int.")
            continue
        if isinstance(d, float):
            if not (_is_int(v) or isinstance(v, float)):
                raise TypeError(f"Config key '{path}' must be a float.")
            continue
        if isinstance(d, str):
            if not isinstance(v, str):
                raise TypeError(f"Config key '{path}' must be a string.")
            continue


def load_config(config_path: str | Path) -> dict[str, Any]:
    cfg_path = Path(config_path).expanduser().resolve()
    if not cfg_path.exists():
        raise FileNotFoundError(f"Config not found: {cfg_path}")
    cfg_dir = cfg_path.parent

    with cfg_path.open("r", encoding="utf-8") as f:
        user_cfg = yaml.safe_load(f) or {}

    _validate_user_config(user_cfg, DEFAULT_CONFIG)
    cfg = _deep_update(DEFAULT_CONFIG, user_cfg)
    _validate_config_types(cfg, DEFAULT_CONFIG)

    cfg["config_path"] = cfg_path
    cfg["config_dir"] = cfg_dir

    inputs = cfg["inputs"]
    outputs = cfg["outputs"]

    inputs["input_catalog"] = _resolve_path(inputs["input_catalog"], cfg_dir)
    inputs["image_list_file"] = _resolve_path(inputs["image_list_file"], cfg_dir)
    inputs["gaiaxp_synphot_csv"] = _resolve_path(inputs["gaiaxp_synphot_csv"], cfg_dir)

    work_dir = _resolve_path(outputs["work_dir"], cfg_dir)
    outputs["work_dir"] = work_dir
    outputs["epsf_dir"] = _resolve_path(outputs["epsf_dir"], work_dir)
    outputs["patches_dir"] = _resolve_path(outputs["patches_dir"], work_dir)
    outputs["patch_inputs_dir"] = _resolve_path(outputs["patch_inputs_dir"], work_dir)
    outputs["tractor_out_dir"] = _resolve_path(outputs["tractor_out_dir"], work_dir)
    outputs["final_catalog"] = _resolve_path(outputs["final_catalog"], work_dir)
    outputs["cropped_images_dir"] = _resolve_path(outputs["cropped_images_dir"], work_dir)

    merge_cfg = cfg["merge"]
    if merge_cfg.get("wcs_fits"):
        merge_cfg["wcs_fits"] = _resolve_path(merge_cfg["wcs_fits"], cfg_dir)

    moffat_cfg = cfg["moffat_psf"]
    if moffat_cfg.get("module_path"):
        moffat_cfg["module_path"] = _resolve_path(moffat_cfg["module_path"], cfg_dir)

    logging_cfg = cfg["logging"]
    if logging_cfg.get("file"):
        logging_cfg["file"] = _resolve_path(logging_cfg["file"], cfg_dir)

    return cfg


def write_sample_config(out_path: str | Path, overwrite: bool = False) -> Path:
    out = Path(out_path).expanduser()
    out_abs = out.resolve()
    sample_path = Path(__file__).resolve().parent / "data" / "sample_config.yaml"
    if not sample_path.exists():
        raise FileNotFoundError(f"Sample config not found: {sample_path}")
    if out_abs.exists() and not overwrite:
        raise FileExistsError(f"Refusing to overwrite existing file: {out_abs}")
    out_abs.parent.mkdir(parents=True, exist_ok=True)
    out_abs.write_text(sample_path.read_text(encoding="utf-8"), encoding="utf-8")
    return out_abs



from __future__ import annotations

import concurrent.futures as cf
import gzip
import json
import logging
import pickle
import sys
import time
from pathlib import Path

import numpy as np
import pandas as pd

logger = logging.getLogger("tract7dt.patch_inputs")


def build_patch_inputs(
    *,
    patch_def_path: Path,
    out_dir: Path,
    image_dict: dict,
    input_catalog: pd.DataFrame,
    include_wcs_in_payload: bool,
    save_patch_inputs: bool,
    skip_empty_patch: bool,
    max_workers: int,
    gzip_compresslevel: int,
    keep_payloads_in_memory: bool,
) -> list[dict]:
    out_dir.mkdir(parents=True, exist_ok=True)
    assert patch_def_path.exists(), f"Missing patch defs: {patch_def_path}"

    patch_defs = json.loads(patch_def_path.read_text())
    logger.info("Loaded patch defs: %d", len(patch_defs))
    logger.info("SKIP_EMPTY_PATCH: %s", skip_empty_patch)
    logger.info(
        "SAVE_PATCH_INPUTS: %s | MAX_WORKERS_PATCH_INPUTS: %d | GZIP_COMPRESSLEVEL: %d",
        save_patch_inputs,
        int(max_workers),
        int(gzip_compresslevel),
    )

    N_TOTAL = len(patch_defs)
    N_DONE = 0
    N_KEEP = 0
    N_SKIP = 0
    PROGRESS_EVERY = max(1, N_TOTAL // 20)
    use_inline_progress = sys.stderr.isatty()
    tstart = time.time()
    last_non_tty_log = 0.0
    inline_width = 160

    def _progress_bar(done: int, total: int, width: int = 24) -> str:
        if total <= 0:
            return "-" * width
        frac = max(0.0, min(1.0, float(done) / float(total)))
        filled = int(round(frac * width))
        return "#" * filled + "-" * (width - filled)

    def _emit_processing_progress(*, force_log: bool = False, queued_writes: int = 0) -> None:
        nonlocal last_non_tty_log
        msg = (
            f"Patch inputs build [{_progress_bar(N_DONE, N_TOTAL)}] "
            f"{N_DONE}/{N_TOTAL} keep={N_KEEP} skip={N_SKIP} queued={queued_writes} "
            f"(total_patches={N_TOTAL})"
        )
        if use_inline_progress:
            sys.stderr.write("\r" + msg.ljust(inline_width))
            sys.stderr.flush()
        else:
            now = time.time()
            if force_log or (now - last_non_tty_log) >= 10.0:
                logger.info(msg)
                last_non_tty_log = now

    def _emit_write_progress(*, done_write: int, total_write: int, force_log: bool = False) -> None:
        nonlocal last_non_tty_log
        elapsed = max(1e-6, time.time() - tstart)
        rate = done_write / elapsed
        msg = (
            f"Patch inputs write [{_progress_bar(done_write, total_write)}] "
            f"{done_write}/{total_write} rate={rate:.2f}/s "
            f"(kept={total_write} total={N_TOTAL})"
        )
        if use_inline_progress:
            sys.stderr.write("\r" + msg.ljust(inline_width))
            sys.stderr.flush()
        else:
            now = time.time()
            if force_log or (now - last_non_tty_log) >= 10.0:
                logger.info(msg)
                last_non_tty_log = now

    cols = list(input_catalog.columns)
    col_map = {str(c).strip().lower(): str(c) for c in cols}
    _ra_col = col_map.get("ra")
    _dec_col = col_map.get("dec")
    if _ra_col is None or _dec_col is None:
        raise ValueError("input_catalog must have RA/DEC columns")
    if _ra_col != "RA" or _dec_col != "DEC":
        logger.warning(
            "input_catalog uses %s/%s columns; interpreting as Right Ascension / Declination. Expected format is RA/DEC.",
            _ra_col,
            _dec_col,
        )

    ra = pd.to_numeric(input_catalog[_ra_col], errors="coerce").to_numpy(dtype=float)
    dec = pd.to_numeric(input_catalog[_dec_col], errors="coerce").to_numpy(dtype=float)

    wcs_white = next(iter(image_dict.values()))["wcs"]

    ok = np.isfinite(ra) & np.isfinite(dec)
    xpix = np.full(len(ra), np.nan, dtype=float)
    ypix = np.full(len(dec), np.nan, dtype=float)
    if np.any(ok):
        xw, yw = wcs_white.all_world2pix(ra[ok], dec[ok], 0)
        xpix[ok] = xw
        ypix[ok] = yw

    cat_global = input_catalog.copy()
    cat_global["x_pix_white"] = xpix
    cat_global["y_pix_white"] = ypix

    patch_inputs = []
    patch_payloads = [] if keep_payloads_in_memory else None

    xg_all = cat_global["x_pix_white"].to_numpy(dtype=float)
    yg_all = cat_global["y_pix_white"].to_numpy(dtype=float)

    _band_items = [(str(band), d) for band, d in image_dict.items()]

    def _build_payload_for_patch(pdef: dict):
        x0b, x1b, y0b, y1b = pdef["x0_base"], pdef["x1_base"], pdef["y0_base"], pdef["y1_base"]
        x0r, x1r, y0r, y1r = pdef["x0_roi"], pdef["x1_roi"], pdef["y0_roi"], pdef["y1_roi"]

        in_base = (
            np.isfinite(xg_all)
            & np.isfinite(yg_all)
            & (xg_all >= x0b)
            & (xg_all < x1b)
            & (yg_all >= y0b)
            & (yg_all < y1b)
        )
        cat_patch = cat_global.loc[in_base].copy().reset_index(drop=True)
        if skip_empty_patch and len(cat_patch) == 0:
            return None

        cat_patch["x_pix_patch"] = cat_patch["x_pix_white"] - float(x0r)
        cat_patch["y_pix_patch"] = cat_patch["y_pix_white"] - float(y0r)

        per_filter_patch = {}
        for band, d in _band_items:
            img_full = d["img_scaled"]
            sig_total = d["sigma_total_scaled"]
            sig_sky = d["sigma_sky_scaled"]
            bad_full = d.get("bad", None)
            if bad_full is None:
                bad_full = ~np.isfinite(img_full)

            entry = dict(
                img_scaled=img_full[y0r:y1r, x0r:x1r],
                sigma_total_scaled=sig_total[y0r:y1r, x0r:x1r],
                sigma_sky_scaled=sig_sky[y0r:y1r, x0r:x1r],
                bad=np.asarray(bad_full[y0r:y1r, x0r:x1r], dtype=bool),
                scale=float(d.get("scale", 1.0)),
                fp=str(d.get("fp", "")),
                x0w=int(x0r),
                x1w=int(x1r),
                y0w=int(y0r),
                y1w=int(y1r),
            )

            if include_wcs_in_payload:
                wcs_patch = d["wcs"].slice((slice(y0r, y1r), slice(x0r, x1r)))
                entry["wcs_patch"] = wcs_patch
                entry["wcs_full"] = d["wcs"]

            per_filter_patch[band] = entry

        meta = dict(
            patch_tag=pdef["patch_tag"],
            epsf_tag=pdef["epsf_tag"],
            epsf_r=int(pdef["epsf_r"]),
            epsf_c=int(pdef["epsf_c"]),
            patch_r=int(pdef["patch_r"]),
            patch_c=int(pdef["patch_c"]),
            halo_pix=int(pdef["halo_pix"]),
            base_bbox=(int(x0b), int(x1b), int(y0b), int(y1b)),
            roi_bbox=(int(x0r), int(x1r), int(y0r), int(y1r)),
            n_sources=int(len(cat_patch)),
        )

        payload = dict(per_filter_patch=per_filter_patch, cat_patch=cat_patch, meta=meta)
        return meta, payload

    def _write_payload(meta: dict, payload: dict):
        outpath = out_dir / f"{meta['patch_tag']}.pkl.gz"
        with gzip.open(outpath, "wb", compresslevel=int(gzip_compresslevel)) as f:
            pickle.dump(payload, f, protocol=pickle.HIGHEST_PROTOCOL)
        return str(outpath)

    if save_patch_inputs and int(max_workers) > 1:
        futs = {}
        with cf.ThreadPoolExecutor(max_workers=int(max_workers)) as ex:
            _emit_processing_progress(force_log=True, queued_writes=0)
            for pdef in patch_defs:
                N_DONE += 1
                built = _build_payload_for_patch(pdef)
                if built is None:
                    N_SKIP += 1
                    if (N_DONE % PROGRESS_EVERY) == 0 or N_DONE == N_TOTAL:
                        _emit_processing_progress(force_log=True, queued_writes=len(futs))
                    continue
                meta, payload = built
                patch_inputs.append(meta)
                if keep_payloads_in_memory:
                    patch_payloads.append(payload)
                N_KEEP += 1
                futs[ex.submit(_write_payload, meta, payload)] = meta["patch_tag"]

                if (N_DONE % PROGRESS_EVERY) == 0 or N_DONE == N_TOTAL:
                    _emit_processing_progress(force_log=True, queued_writes=len(futs))

            ndone_w = 0
            pending = set(futs.keys())
            total_writes = len(pending)
            _emit_write_progress(done_write=ndone_w, total_write=total_writes, force_log=True)
            while pending:
                done_now, pending = cf.wait(pending, timeout=0.5, return_when=cf.FIRST_COMPLETED)
                if not done_now:
                    _emit_write_progress(done_write=ndone_w, total_write=total_writes, force_log=False)
                    continue
                for fut in done_now:
                    _ = fut.result()
                    ndone_w += 1
                _emit_write_progress(done_write=ndone_w, total_write=total_writes, force_log=True)
    else:
        _emit_processing_progress(force_log=True, queued_writes=0)
        for pdef in patch_defs:
            N_DONE += 1
            built = _build_payload_for_patch(pdef)
            if built is None:
                N_SKIP += 1
                if (N_DONE % PROGRESS_EVERY) == 0 or N_DONE == N_TOTAL:
                    _emit_processing_progress(force_log=True, queued_writes=0)
                continue

            meta, payload = built
            patch_inputs.append(meta)
            if keep_payloads_in_memory:
                patch_payloads.append(payload)
            N_KEEP += 1

            if save_patch_inputs:
                _ = _write_payload(meta, payload)

            if (N_DONE % PROGRESS_EVERY) == 0 or N_DONE == N_TOTAL:
                _emit_processing_progress(force_log=True, queued_writes=0)

    if use_inline_progress:
        sys.stderr.write("\n")
        sys.stderr.flush()

    logger.info("Patch inputs created: %d", len(patch_inputs))
    if save_patch_inputs:
        logger.info("Saved to: %s", out_dir)
    else:
        logger.info("Saved to: (skipped)")

    return patch_inputs



"""
Patch-level Tractor runner for tract7dt.
"""

from __future__ import annotations

import argparse
import gzip
import json
import logging
import math
import os
import pickle
import re
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional

import numpy as np
import pandas as pd

import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt
from matplotlib.cm import ScalarMappable
from matplotlib.colors import Normalize

import sep
from astropy.io import fits

from tractor import ConstantSky, Fluxes, FluxesPhotoCal, Image, NullWCS, PixPos, PointSource, Tractor
from tractor.constrained_optimizer import ConstrainedOptimizer
from tractor.ellipses import EllipseESoft
from tractor.galaxy import DevGalaxy, ExpGalaxy
from tractor.psf import GaussianMixturePSF, HybridPixelizedPSF, NCircularGaussianPSF, PixelizedPSF
from tractor.sersic import SersicGalaxy, SersicIndex
from .moffat_psf import MoffatPSF

sep.set_sub_object_limit(200000)
sep.set_extract_pixstack(1000000)


@dataclass
class PatchRunConfig:
    # Optimization
    n_opt_iters: int = 200
    dlnp_stop: float = 1e-6
    # Flag "non-converged" if we effectively hit max iters.
    # If not converged and niters >= (n_opt_iters - margin), set opt_hit_max_iters=True.
    flag_maxiter_margin: int = 5

    # Catalog init
    r_ap: float = 5.0
    eps_flux: float = 1e-4
    gal_model: str = "dev"  # dev|exp|sersic|star
    sersic_n_init: float = 3.0
    re_fallback_pix: float = 3.0

    # Output cutouts
    save_cutouts: bool = True
    cutout_size_pix: int = 100
    cutout_max_sources: Optional[int] = None
    cutout_start: int = 0
    cutout_data_p_lo: float = 1
    cutout_data_p_hi: float = 99
    cutout_model_p_lo: float = 1
    cutout_model_p_hi: float = 99
    cutout_resid_p_abs: float = 99

    # Per-patch overview
    save_patch_overview: bool = True

    # PSF choices:
    # psf_model: used when ePSF exists
    # psf_fallback_model: used when ePSF is missing
    psf_model: str = "hybrid"  # hybrid|pixelized|gaussian_mixture_from_stamp
    psf_fallback_model: str = "gaussian_mixture"  # gaussian_mixture|ncircular_gaussian|moffat
    min_epsf_nstars_for_use: int = 5
    ncircular_gaussian_n: int = 3  # allowed: 1|2|3
    hybrid_psf_n: int = 4
    gaussian_mixture_n: int = 4
    fallback_default_fwhm_pix: float = 4.0
    fallback_stamp_radius_pix: float = 25.0
    moffat_beta: float = 3.5
    moffat_radius_pix: float = 25.0


def _ensure_dir(p: Path) -> None:
    p.mkdir(parents=True, exist_ok=True)


def _make_logger(outdir: Path, tag: str) -> logging.Logger:
    _ensure_dir(outdir)
    log = logging.getLogger(f"tractor_patch.{tag}")
    log.setLevel(logging.INFO)
    log.propagate = False
    if not log.handlers:
        fmt = logging.Formatter("%(asctime)s | %(levelname)s | %(message)s")
        fh = logging.FileHandler(outdir / f"{tag}.log", mode="w")
        fh.setFormatter(fmt)
        fh.setLevel(logging.INFO)
        log.addHandler(fh)
        sh = logging.StreamHandler(sys.stdout)
        sh.setFormatter(fmt)
        sh.setLevel(logging.INFO)
        log.addHandler(sh)
    return log


def _patch_tag_from_meta(meta: dict | None) -> str:
    if not meta:
        return "patch"
    t = meta.get("patch_tag", None)
    if isinstance(t, str) and t.strip():
        return t.strip()
    return "patch"


def _roi_bbox_from_meta(meta: dict) -> tuple[int, int, int, int]:
    if "roi_bbox" in meta:
        b = meta["roi_bbox"]
        if isinstance(b, (list, tuple)) and len(b) == 4:
            return (int(b[0]), int(b[1]), int(b[2]), int(b[3]))
    keys = ("x0_roi", "x1_roi", "y0_roi", "y1_roi")
    if all(k in meta for k in keys):
        return (int(meta["x0_roi"]), int(meta["x1_roi"]), int(meta["y0_roi"]), int(meta["y1_roi"]))
    raise ValueError("meta must contain roi_bbox or (x0_roi,x1_roi,y0_roi,y1_roi)")


def load_patch_inputs(pkl_gz_path: str | Path) -> dict:
    p = Path(pkl_gz_path)
    with gzip.open(p, "rb") as f:
        payload = pickle.load(f)
    if not isinstance(payload, dict):
        raise ValueError("Patch input payload must be a dict")
    return payload


def _fallback_fwhm_from_header(
    fp: str,
    *,
    default_fwhm_pix: float,
    log: logging.Logger | None = None,
) -> float:
    fwhm_pix = float(default_fwhm_pix)
    used_default = True
    try:
        with fits.open(fp, memmap=True) as hdul:
            hdr = hdul[0].header
        v = hdr.get("PEEING", None)
        if v is None:
            v = hdr.get("SEEING", None)
        if v is not None and np.isfinite(v) and v > 0:
            fwhm_pix = float(v)
            used_default = False
    except Exception:
        pass
    if used_default and log is not None:
        log.info("PEEING/SEEING missing/invalid in %s; using default_fwhm_pix=%s", fp, str(default_fwhm_pix))
    return max(float(fwhm_pix), 1.0)


def _gaussian_stamp_from_fwhm(*, fwhm_pix: float, radius_pix: float) -> np.ndarray:
    sigma = max(float(fwhm_pix) / 2.3548200450309493, 1e-3)
    r = max(1, int(np.ceil(float(radius_pix))))
    yy, xx = np.mgrid[-r : r + 1, -r : r + 1].astype(np.float32)
    rr2 = (xx * xx + yy * yy).astype(np.float64)
    img = np.exp(-0.5 * rr2 / (sigma * sigma))
    s = float(np.nansum(img))
    if not np.isfinite(s) or s <= 0:
        raise RuntimeError("Failed to build Gaussian fallback stamp.")
    return (img / s).astype(np.float32)


def _center_crop_or_pad(arr: np.ndarray, shape: tuple[int, int]) -> np.ndarray:
    out = np.zeros(shape, dtype=float)
    y_in, x_in = arr.shape
    y_out, x_out = shape
    y0_in = max(0, (y_in - y_out) // 2)
    x0_in = max(0, (x_in - x_out) // 2)
    y0_out = max(0, (y_out - y_in) // 2)
    x0_out = max(0, (x_out - x_in) // 2)
    y1_in = y0_in + min(y_in, y_out)
    x1_in = x0_in + min(x_in, x_out)
    y1_out = y0_out + min(y_in, y_out)
    x1_out = x0_out + min(x_in, x_out)
    out[y0_out:y1_out, x0_out:x1_out] = arr[y0_in:y1_in, x0_in:x1_in]
    return out


def _psf_to_array(psf: Any, shape: tuple[int, int]) -> np.ndarray | None:
    for kwargs in ({}, {"stamp": int(shape[0])}, {"size": int(shape[0])}, {"halfsize": int(shape[0] // 2)}):
        try:
            patch = psf.getPointSourcePatch(0.0, 0.0, **kwargs)
            if patch is None:
                continue
            if hasattr(patch, "getImage"):
                arr = np.asarray(patch.getImage(), dtype=float)
            elif hasattr(patch, "image"):
                arr = np.asarray(patch.image, dtype=float)
            else:
                arr = np.asarray(patch, dtype=float)
            if arr.size:
                return _center_crop_or_pad(arr, shape)
        except Exception:
            continue
    if hasattr(psf, "getImage"):
        try:
            arr = np.asarray(psf.getImage(), dtype=float)
            if arr.size:
                return _center_crop_or_pad(arr, shape)
        except Exception:
            pass
    if hasattr(psf, "pix"):
        try:
            arr = np.asarray(psf.pix.getImage(), dtype=float)
            if arr.size:
                return _center_crop_or_pad(arr, shape)
        except Exception:
            pass
    return None


def _save_hybrid_psf_diagnostics(
    *,
    epsf_arr: np.ndarray,
    hybrid_psf: HybridPixelizedPSF,
    outpath: Path,
    log: logging.Logger,
) -> None:
    gauss_arr = _psf_to_array(hybrid_psf.gauss, epsf_arr.shape)
    if gauss_arr is not None:
        gauss_arr = np.nan_to_num(gauss_arr, nan=0.0, posinf=0.0, neginf=0.0)
        gs = float(np.sum(gauss_arr))
        if gs > 0:
            gauss_arr = gauss_arr / gs

    gauss_resid = (epsf_arr - gauss_arr) if gauss_arr is not None else None

    dvmin, dvmax = _robust_limits_percentile(epsf_arr, p_lo=1, p_hi=99)
    gvmin, gvmax = _robust_limits_percentile(gauss_arr, p_lo=1, p_hi=99) if gauss_arr is not None else (None, None)
    grvmin, grvmax = (
        _robust_resid_limits_percentile(gauss_resid, p_abs=99)
        if gauss_resid is not None
        else (None, None)
    )

    if gauss_arr is None:
        fig, axes = plt.subplots(1, 1, figsize=(4, 4), constrained_layout=True)
        panels = [
            ("ePSF (normalized)", epsf_arr, dvmin, dvmax, "ePSF"),
        ]
    else:
        fig, axes = plt.subplots(1, 3, figsize=(12, 4), constrained_layout=True)
        panels = [
            ("ePSF (normalized)", epsf_arr, dvmin, dvmax, "ePSF"),
            ("Gaussian model (normalized)", gauss_arr, gvmin, gvmax, "Gauss"),
            ("ePSF - gauss", gauss_resid, grvmin, grvmax, "Gauss resid"),
        ]
    for ax, (title, arr, vmin, vmax, cblab) in zip(axes, panels):
        im = ax.imshow(arr, origin="lower", cmap="gray", vmin=vmin, vmax=vmax, interpolation="nearest")
        ax.set_title(title, fontsize=10)
        ax.set_xticks([])
        ax.set_yticks([])
        amin, amax = _true_minmax(arr)
        if amin is not None and amax is not None:
            sm = ScalarMappable(norm=Normalize(vmin=amin, vmax=amax), cmap=im.cmap)
            sm.set_array([])
            cb = fig.colorbar(sm, ax=ax, fraction=0.046, pad=0.02, extend="neither")
            cb.set_label(cblab)
    outpath.parent.mkdir(parents=True, exist_ok=True)
    plt.savefig(outpath, dpi=120)
    plt.close(fig)
    log.info("[PSF] Wrote hybrid PSF diagnostics: %s", str(outpath))


def _save_gaussian_mixture_psf_diagnostics(
    *,
    epsf_arr: np.ndarray,
    gauss_psf: GaussianMixturePSF,
    outpath: Path,
    log: logging.Logger,
) -> None:
    gauss_arr = _psf_to_array(gauss_psf, epsf_arr.shape)
    if gauss_arr is None:
        raise RuntimeError("Could not render GaussianMixturePSF to array for diagnostics.")
    gauss_arr = np.nan_to_num(gauss_arr, nan=0.0, posinf=0.0, neginf=0.0)
    gs = float(np.sum(gauss_arr))
    if gs > 0:
        gauss_arr = gauss_arr / gs

    resid = epsf_arr - gauss_arr

    dvmin, dvmax = _robust_limits_percentile(epsf_arr, p_lo=1, p_hi=99)
    gvmin, gvmax = _robust_limits_percentile(gauss_arr, p_lo=1, p_hi=99)
    rvmin, rvmax = _robust_resid_limits_percentile(resid, p_abs=99)

    fig, axes = plt.subplots(1, 3, figsize=(12, 4), constrained_layout=True)
    panels = [
        ("ePSF (normalized)", epsf_arr, dvmin, dvmax, "ePSF"),
        ("GaussianMixture model (normalized)", gauss_arr, gvmin, gvmax, "Gauss"),
        ("ePSF - gauss", resid, rvmin, rvmax, "Gauss resid"),
    ]
    for ax, (title, arr, vmin, vmax, cblab) in zip(axes, panels):
        im = ax.imshow(arr, origin="lower", cmap="gray", vmin=vmin, vmax=vmax, interpolation="nearest")
        ax.set_title(title, fontsize=10)
        ax.set_xticks([])
        ax.set_yticks([])
        amin, amax = _true_minmax(arr)
        if amin is not None and amax is not None:
            sm = ScalarMappable(norm=Normalize(vmin=amin, vmax=amax), cmap=im.cmap)
            sm.set_array([])
            cb = fig.colorbar(sm, ax=ax, fraction=0.046, pad=0.02, extend="neither")
            cb.set_label(cblab)
    outpath.parent.mkdir(parents=True, exist_ok=True)
    plt.savefig(outpath, dpi=120)
    plt.close(fig)
    log.info("[PSF] Wrote GaussianMixture PSF diagnostics: %s", str(outpath))


def build_psf_by_band_for_patch(
    *,
    per_filter_patch: dict,
    epsf_root: Path,
    epsf_tag: str,
    cfg: PatchRunConfig,
    log: logging.Logger,
) -> tuple[dict[str, Any], dict[str, dict[str, Any]]]:
    psf_by_band: dict[str, Any] = {}
    psf_audit: dict[str, dict[str, Any]] = {}
    for band, d in per_filter_patch.items():
        band = str(band)
        epsf_path = epsf_root / band / str(epsf_tag) / "epsf.npy"
        band_audit: dict[str, Any] = dict(
            epsf_path=str(epsf_path),
            epsf_exists=bool(epsf_path.exists()),
            used_epsf=False,
            fallback_reason=None,
            epsf_nstars_used=None,
            min_epsf_nstars_for_use=int(max(0, int(getattr(cfg, "min_epsf_nstars_for_use", 5)))),
            psf_model=None,
        )
        use_epsf = epsf_path.exists()
        fallback_reason = "missing ePSF"
        if use_epsf:
            min_n = max(0, int(getattr(cfg, "min_epsf_nstars_for_use", 5)))
            if min_n > 0:
                meta_path = epsf_path.parent / "meta.json"
                nstars_used = None
                try:
                    if meta_path.exists():
                        meta = json.loads(meta_path.read_text())
                        v = meta.get("nstars_used", None)
                        if v is not None:
                            nstars_used = int(v)
                            band_audit["epsf_nstars_used"] = int(nstars_used)
                except Exception as e:
                    log.warning("[PSF] %s: failed to read ePSF meta %s: %s", band, str(meta_path), str(e))
                if nstars_used is not None and nstars_used < min_n:
                    use_epsf = False
                    fallback_reason = (
                        f"low-star ePSF (nstars_used={nstars_used} < min_epsf_nstars_for_use={min_n})"
                    )
                    log.warning(
                        "[PSF] %s: ePSF exists at %s but %s; using fallback path",
                        band,
                        str(epsf_path),
                        fallback_reason,
                    )

        if use_epsf:
            epsf_arr = np.load(epsf_path)
            epsf_arr = np.asarray(epsf_arr, dtype=np.float32)
            epsf_arr = np.nan_to_num(epsf_arr, nan=0.0, posinf=0.0, neginf=0.0)
            s = float(np.sum(epsf_arr))
            if s > 0:
                epsf_arr = epsf_arr / s
            pixpsf = PixelizedPSF(epsf_arr)
            psf_model = str(getattr(cfg, "psf_model", "hybrid")).lower()
            band_audit["used_epsf"] = True
            band_audit["psf_model"] = str(psf_model)
            if psf_model == "pixelized":
                psf_by_band[band] = pixpsf
            elif psf_model == "gaussian_mixture_from_stamp":
                gauss = GaussianMixturePSF.fromStamp(epsf_arr, N=max(1, int(cfg.gaussian_mixture_n)))
                psf_by_band[band] = gauss
                try:
                    _save_gaussian_mixture_psf_diagnostics(
                        epsf_arr=epsf_arr,
                        gauss_psf=gauss,
                        outpath=epsf_path.parent / "gaussian_mixture_psf_diagnostics.png",
                        log=log,
                    )
                except Exception as e:
                    log.warning("[PSF] GaussianMixture PSF diagnostics failed: %s", str(e))
            else:
                hybrid = HybridPixelizedPSF(pixpsf, gauss=None, N=int(cfg.hybrid_psf_n), cx=0.0, cy=0.0)
                psf_by_band[band] = hybrid
                try:
                    _save_hybrid_psf_diagnostics(
                        epsf_arr=epsf_arr,
                        hybrid_psf=hybrid,
                        outpath=epsf_path.parent / "hybrid_psf_diagnostics.png",
                        log=log,
                    )
                except Exception as e:
                    log.warning("[PSF] Hybrid PSF diagnostics failed: %s", str(e))
        else:
            fp = d.get("fp", None)
            if fp:
                fwhm_pix = _fallback_fwhm_from_header(
                    str(fp),
                    default_fwhm_pix=cfg.fallback_default_fwhm_pix,
                    log=log,
                )
            else:
                fwhm_pix = float(cfg.fallback_default_fwhm_pix)
            psf_fallback_model = str(getattr(cfg, "psf_fallback_model", "gaussian_mixture")).lower()
            band_audit["used_epsf"] = False
            band_audit["fallback_reason"] = str(fallback_reason)
            band_audit["psf_model"] = str(psf_fallback_model)
            if psf_fallback_model == "ncircular_gaussian":
                sigma = max(float(fwhm_pix) / 2.3548200450309493, 1e-3)
                ncomp = int(getattr(cfg, "ncircular_gaussian_n", 3))
                if ncomp == 1:
                    sigmas = [sigma]
                    weights = [1.0]
                elif ncomp == 2:
                    sigmas = [sigma, 2.5 * sigma]
                    weights = [0.85, 0.15]
                elif ncomp == 3:
                    sigmas = [sigma, 2.0 * sigma, 4.0 * sigma]
                    weights = [0.8, 0.15, 0.05]
                else:
                    raise ValueError(f"ncircular_gaussian_n must be 1, 2, or 3. Got: {ncomp}")
                psf_by_band[band] = NCircularGaussianPSF(sigmas, weights)
            elif psf_fallback_model == "moffat":
                psf_by_band[band] = MoffatPSF(
                    fwhm_pix=float(fwhm_pix),
                    beta=float(cfg.moffat_beta),
                    radius_pix=float(cfg.moffat_radius_pix),
                )
            else:
                stamp = _gaussian_stamp_from_fwhm(fwhm_pix=float(fwhm_pix), radius_pix=float(cfg.fallback_stamp_radius_pix))
                psf_by_band[band] = GaussianMixturePSF.fromStamp(stamp, N=max(1, int(cfg.gaussian_mixture_n)))
            log.info("[PSF] %s: %s at %s -> fallback %s", band, fallback_reason, str(epsf_path), str(psf_by_band[band]))
        psf_audit[band] = band_audit
    return psf_by_band, psf_audit


def build_tractor_images_from_per_filter_patch(
    *,
    per_filter_patch: dict,
    psf_by_filt: dict[str, Any],
    use_sky_sigma: bool = True,
) -> tuple[list[Image], list[dict[str, Any]]]:
    tractor_images: list[Image] = []
    image_data: list[dict[str, Any]] = []
    for filt, d in per_filter_patch.items():
        filt = str(filt)
        img = np.ascontiguousarray(d["img_scaled"], dtype=np.float32)
        sig = d["sigma_sky_scaled"] if use_sky_sigma else d["sigma_total_scaled"]
        sig = np.ascontiguousarray(sig, dtype=np.float32)
        bad = np.asarray(d.get("bad", ~np.isfinite(img)), dtype=bool)
        bad |= ~np.isfinite(img) | ~np.isfinite(sig) | (sig <= 0)

        invvar = np.zeros_like(img, dtype=np.float32)
        good = ~bad
        invvar[good] = 1.0 / np.maximum(sig[good] * sig[good], 1e-12)

        tim = Image(
            data=img,
            invvar=invvar,
            psf=psf_by_filt.get(filt, None),
            wcs=NullWCS(),
            sky=ConstantSky(0.0),
            photocal=FluxesPhotoCal(str(filt)),
            name=str(filt),
        )
        tractor_images.append(tim)
        image_data.append(
            dict(
                filter=str(filt),
                fp=d.get("fp", None),
                scale=float(d.get("scale", np.nan)) if d.get("scale", None) is not None else np.nan,
                data=img,
                invvar=invvar,
                sigma_used=sig,
                bad=bad,
            )
        )
    return tractor_images, image_data


def build_catalog_from_cat_patch(cat_patch: pd.DataFrame, tractor_images: list[Image], cfg: PatchRunConfig) -> list[Any]:
    if "x_pix_patch" not in cat_patch.columns or "y_pix_patch" not in cat_patch.columns:
        raise ValueError("cat_patch must contain columns: x_pix_patch, y_pix_patch")
    xs = np.asarray(cat_patch["x_pix_patch"], dtype=np.float32)
    ys = np.asarray(cat_patch["y_pix_patch"], dtype=np.float32)

    bandnames = [str(getattr(tim, "name", f"band{i}")) for i, tim in enumerate(tractor_images)]
    col_map = {str(c).strip().lower(): str(c) for c in cat_patch.columns}
    init_fluxes = [dict((bn, float(cfg.eps_flux)) for bn in bandnames) for _ in range(len(cat_patch))]

    for tim, bn in zip(tractor_images, bandnames):
        img = np.ascontiguousarray(tim.getImage(), dtype=np.float32)
        inv = np.ascontiguousarray(tim.getInvvar(), dtype=np.float32)
        good = np.isfinite(img) & np.isfinite(inv) & (inv > 0)
        err = np.full_like(img, np.inf, dtype=np.float32)
        err[good] = (1.0 / np.sqrt(inv[good])).astype(np.float32)
        mask = ~good
        f_ap, _ferr, _flag = sep.sum_circle(img, xs, ys, r=float(cfg.r_ap), err=err, mask=mask)
        f_ap = np.asarray(f_ap, dtype=np.float32)
        f_ap = np.where(np.isfinite(f_ap), f_ap, 0.0)

        flux_col = col_map.get(f"flux_{bn}".lower())
        if flux_col is not None:
            f_cat = pd.to_numeric(cat_patch[flux_col], errors="coerce").to_numpy(dtype=np.float32)
            use_cat = np.isfinite(f_cat) & (f_cat > 0)
            f = np.where(use_cat, f_cat, f_ap).astype(np.float32)
        else:
            f = f_ap
        f = np.maximum(f, float(cfg.eps_flux))
        for i in range(len(cat_patch)):
            init_fluxes[i][bn] = float(f[i])

    if "TYPE" in cat_patch.columns:
        t = cat_patch["TYPE"].astype(str).str.upper().to_numpy()
    else:
        t = np.full(len(cat_patch), str(cfg.gal_model).upper(), dtype=object)

    if "ELL" in cat_patch.columns:
        ell = pd.to_numeric(cat_patch["ELL"], errors="coerce").to_numpy(dtype=float)
        ab = np.clip(1.0 - np.nan_to_num(ell, nan=0.0), 0.1, 1.0)
    else:
        ab = np.full(len(cat_patch), 0.8, dtype=float)

    if "THETA" in cat_patch.columns:
        phi = pd.to_numeric(cat_patch["THETA"], errors="coerce").to_numpy(dtype=float)
        phi = np.nan_to_num(phi, nan=0.0)
        phi = -np.where(phi >= 0, 90 - phi, -90 - phi)
    else:
        phi = np.zeros(len(cat_patch), dtype=float)

    if "Re" in cat_patch.columns:
        re_ = pd.to_numeric(cat_patch["Re"], errors="coerce").to_numpy(dtype=float)
        re_ = np.where(np.isfinite(re_), re_, float(cfg.re_fallback_pix))
        re_ = np.clip(re_, 0.3, 100.0)
    else:
        re_ = np.full(len(cat_patch), float(cfg.re_fallback_pix), dtype=float)

    catalog: list[Any] = []
    for i in range(len(cat_patch)):
        pos = PixPos(float(xs[i]), float(ys[i]))
        bright = Fluxes(**init_fluxes[i])
        ti = str(t[i]).upper()
        if ti == "STAR":
            src = PointSource(pos, bright)
        else:
            RE_MIN = 0.1
            RE_MAX = 1000.0
            EE_MAX = 100.0
            shape = EllipseESoft.fromRAbPhi(float(re_[i]), float(ab[i]), float(phi[i]))
            shape.lowers = [float(np.log(RE_MIN)), -float(EE_MAX), -float(EE_MAX)]
            shape.uppers = [float(np.log(RE_MAX)), float(EE_MAX), float(EE_MAX)]

            if ti == "SERSIC":
                src = SersicGalaxy(pos, bright, shape, SersicIndex(float(cfg.sersic_n_init)))
            elif ti == "DEV":
                src = DevGalaxy(pos, bright, shape)
            elif ti == "EXP":
                src = ExpGalaxy(pos, bright, shape)
            else:
                gm = str(cfg.gal_model).lower()
                if gm == "star":
                    src = PointSource(pos, bright)
                elif gm == "sersic":
                    src = SersicGalaxy(pos, bright, shape, SersicIndex(float(cfg.sersic_n_init)))
                elif gm == "dev":
                    src = DevGalaxy(pos, bright, shape)
                else:
                    src = ExpGalaxy(pos, bright, shape)
        catalog.append(src)
    return catalog


def _robust_limits_percentile(arr: np.ndarray, p_lo: float, p_hi: float, min_span: float = 1e-6):
    a = np.asarray(arr, dtype=float)
    a = a[np.isfinite(a)]
    if a.size == 0:
        return None, None
    vmin, vmax = np.percentile(a, [p_lo, p_hi])
    vmin, vmax = float(vmin), float(vmax)
    if not (np.isfinite(vmin) and np.isfinite(vmax)):
        return None, None
    if (vmax - vmin) < min_span:
        vmin = float(np.min(a))
        vmax = float(np.max(a))
        if (vmax - vmin) < min_span:
            vmin -= 0.5
            vmax += 0.5
    return vmin, vmax


def _robust_resid_limits_percentile(resid: np.ndarray, p_abs: float = 99, min_span: float = 1e-6):
    r = np.asarray(resid, dtype=float)
    r = r[np.isfinite(r)]
    if r.size == 0:
        return None, None
    a = np.nanpercentile(np.abs(r), p_abs)
    if not np.isfinite(a) or a < min_span:
        a = np.nanmax(np.abs(r))
        if not np.isfinite(a) or a < min_span:
            a = 1.0
    return float(-a), float(+a)


def _true_minmax(arr: np.ndarray):
    a = np.asarray(arr, dtype=float)
    a = a[np.isfinite(a)]
    if a.size == 0:
        return None, None
    return float(a.min()), float(a.max())


def _white_from_images(*, images: list[Image], model: bool, tr: Optional[Tractor] = None) -> np.ndarray:
    num = None
    den = None
    for i, tim in enumerate(images):
        img = np.asarray(tim.getImage(), dtype=np.float32)
        inv = np.asarray(tim.getInvvar(), dtype=np.float32)
        good = np.isfinite(img) & np.isfinite(inv) & (inv > 0)
        w = np.where(good, inv, 0.0).astype(np.float32, copy=False)
        if model:
            if tr is None:
                raise ValueError("tr is required when model=True")
            m = np.asarray(tr.getModelImage(i), dtype=np.float32)
            sky = float(getattr(getattr(tim, "sky", None), "getValue", lambda: 0.0)())
            img_use = m + sky
        else:
            img_use = img
        if num is None:
            num = w * img_use
            den = w.copy()
        else:
            num += w * img_use
            den += w
    if num is None or den is None:
        raise ValueError("No images provided for white")
    out = np.full_like(den, np.nan, dtype=np.float32)
    ok = den > 0
    out[ok] = (num[ok] / den[ok]).astype(np.float32)
    return out


def save_patch_overview(
    *,
    tr: Tractor,
    cat_patch: Optional[pd.DataFrame],
    meta: Optional[dict],
    outpath: Path,
    data_p_lo: float,
    data_p_hi: float,
    model_p_lo: float,
    model_p_hi: float,
    resid_p_abs: float,
    log: Optional[logging.Logger] = None,
) -> None:
    meta = meta or {}
    patch_tag = str(meta.get("patch_tag", "patch"))
    nsrc = int(meta.get("n_sources", len(tr.catalog))) if hasattr(tr, "catalog") else 0

    white_data = _white_from_images(images=list(tr.images), model=False, tr=None)
    white_model = _white_from_images(images=list(tr.images), model=True, tr=tr)
    white_resid = np.asarray(white_data - white_model, dtype=np.float32)

    dvmin, dvmax = _robust_limits_percentile(white_data, p_lo=data_p_lo, p_hi=data_p_hi)
    mvmin, mvmax = _robust_limits_percentile(white_model, p_lo=model_p_lo, p_hi=model_p_hi)
    rvmin, rvmax = _robust_resid_limits_percentile(white_resid, p_abs=resid_p_abs)

    fig, axes = plt.subplots(1, 3, figsize=(12, 4), constrained_layout=True)
    panels = [
        ("Data (white)", white_data, dvmin, dvmax, "Data"),
        ("Model+Sky (white)", white_model, mvmin, mvmax, "Model+Sky"),
        ("Residual (white)", white_resid, rvmin, rvmax, "Residual"),
    ]

    x_fit = np.array([float((s.getPosition() if hasattr(s, "getPosition") else s.pos).x) for s in tr.catalog], dtype=float)
    y_fit = np.array([float((s.getPosition() if hasattr(s, "getPosition") else s.pos).y) for s in tr.catalog], dtype=float)
    x_orig = y_orig = None
    if isinstance(cat_patch, pd.DataFrame) and "x_pix_patch" in cat_patch.columns and "y_pix_patch" in cat_patch.columns:
        x_orig = pd.to_numeric(cat_patch["x_pix_patch"], errors="coerce").to_numpy(dtype=float)
        y_orig = pd.to_numeric(cat_patch["y_pix_patch"], errors="coerce").to_numpy(dtype=float)

    for ax, (title, arr, vmin, vmax, cblab) in zip(axes, panels):
        im = ax.imshow(arr, origin="lower", cmap="gray", vmin=vmin, vmax=vmax, interpolation="nearest")
        ax.set_title(title, fontsize=10)
        ax.set_xticks([])
        ax.set_yticks([])
        if x_orig is not None and y_orig is not None:
            ok = np.isfinite(x_orig) & np.isfinite(y_orig)
            if np.any(ok):
                ax.scatter(x_orig[ok], y_orig[ok], s=18, marker="x", c="orange", linewidths=0.8, alpha=0.8)
        okf = np.isfinite(x_fit) & np.isfinite(y_fit)
        if np.any(okf):
            ax.scatter(x_fit[okf], y_fit[okf], s=18, marker="x", c="deepskyblue", linewidths=0.8, alpha=0.8)
        amin, amax = _true_minmax(arr)
        if amin is not None and amax is not None:
            sm = ScalarMappable(norm=Normalize(vmin=amin, vmax=amax), cmap=im.cmap)
            sm.set_array([])
            cb = fig.colorbar(sm, ax=ax, fraction=0.046, pad=0.02, extend="neither")
            cb.set_label(cblab)

    fig.suptitle(f"patch={patch_tag} | n_sources={nsrc}\norig=x orange, fit=x deepskyblue", fontsize=12)
    outpath.parent.mkdir(parents=True, exist_ok=True)
    plt.savefig(outpath, dpi=120)
    plt.close(fig)
    if log:
        log.info("Wrote patch overview: %s", str(outpath))


def _cut(arr: np.ndarray, cx: float, cy: float, size: int):
    ix = int(round(float(cx)))
    iy = int(round(float(cy)))
    half = max(1, int(size) // 2)
    ny, nx = arr.shape
    ix = int(np.clip(ix, 0, nx - 1))
    iy = int(np.clip(iy, 0, ny - 1))

    hx = min(half, ix, (nx - 1 - ix))
    hy = min(half, iy, (ny - 1 - iy))

    x0 = ix - hx
    x1 = ix + hx + 1
    y0 = iy - hy
    y1 = iy + hy + 1
    return arr[y0:y1, x0:x1], (y0, y1, x0, x1)


def save_source_montages(
    *,
    tr: Tractor,
    cat_patch: Optional[pd.DataFrame] = None,
    meta: Optional[dict] = None,
    outdir: Path,
    size: int,
    per_band_scale: bool,
    max_sources: Optional[int],
    start: int,
    data_p_lo: float,
    data_p_hi: float,
    model_p_lo: float,
    model_p_hi: float,
    resid_p_abs: float,
    log: Optional[logging.Logger] = None,
) -> None:
    _ensure_dir(outdir)
    meta = meta or {}
    patch_tag = str(meta.get("patch_tag", "patch"))
    roi_bbox = meta.get("roi_bbox", None)
    x0_roi = y0_roi = None
    if isinstance(roi_bbox, (list, tuple)) and len(roi_bbox) == 4:
        try:
            x0_roi = float(roi_bbox[0])
            y0_roi = float(roi_bbox[2])
        except Exception:
            x0_roi = y0_roi = None

    ids = None
    types = None
    x_orig_all = y_orig_all = None
    xw_orig_all = yw_orig_all = None
    if isinstance(cat_patch, pd.DataFrame):
        if "ID" in cat_patch.columns:
            ids = cat_patch["ID"].astype(str).to_numpy()
        if "TYPE" in cat_patch.columns:
            types = cat_patch["TYPE"].astype(str).to_numpy()
        if "x_pix_patch" in cat_patch.columns and "y_pix_patch" in cat_patch.columns:
            x_orig_all = pd.to_numeric(cat_patch["x_pix_patch"], errors="coerce").to_numpy(dtype=float)
            y_orig_all = pd.to_numeric(cat_patch["y_pix_patch"], errors="coerce").to_numpy(dtype=float)
        if "x_pix_white" in cat_patch.columns and "y_pix_white" in cat_patch.columns:
            xw_orig_all = pd.to_numeric(cat_patch["x_pix_white"], errors="coerce").to_numpy(dtype=float)
            yw_orig_all = pd.to_numeric(cat_patch["y_pix_white"], errors="coerce").to_numpy(dtype=float)

    bandnames = [str(getattr(tim, "name", f"band{i}")) for i, tim in enumerate(tr.images)]

    data_by_band = [np.asarray(tim.getImage(), dtype=np.float32) for tim in tr.images]
    model_by_band = [np.asarray(tr.getModelImage(i), dtype=np.float32) for i in range(len(tr.images))]
    sky_by_band = [float(getattr(getattr(tim, "sky", None), "getValue", lambda: 0.0)()) for tim in tr.images]

    indices = list(range(len(tr.catalog)))
    indices = [i for i in indices if i >= int(start)]
    if max_sources is not None:
        indices = indices[: int(max_sources)]

    if log:
        log.info("Saving %d montage(s) to %s", len(indices), str(outdir))

    for si in indices:
        src = tr.catalog[si]
        pos = src.getPosition() if hasattr(src, "getPosition") else src.pos
        cx, cy = float(pos.x), float(pos.y)

        sid = None
        if ids is not None and si < len(ids):
            sid = str(ids[si])
        stype = None
        if types is not None and si < len(types):
            stype = str(types[si])

        x0_fit_white = None
        y0_fit_white = None
        if x0_roi is not None and y0_roi is not None and np.isfinite(cx) and np.isfinite(cy):
            x0_fit_white = float(cx + x0_roi)
            y0_fit_white = float(cy + y0_roi)

        x0_orig = y0_orig = None
        if x_orig_all is not None and y_orig_all is not None and si < len(x_orig_all):
            x0_orig = float(x_orig_all[si]) if np.isfinite(x_orig_all[si]) else None
            y0_orig = float(y_orig_all[si]) if np.isfinite(y_orig_all[si]) else None

        x0_orig_white = y0_orig_white = None
        if xw_orig_all is not None and yw_orig_all is not None and si < len(xw_orig_all):
            x0_orig_white = float(xw_orig_all[si]) if np.isfinite(xw_orig_all[si]) else None
            y0_orig_white = float(yw_orig_all[si]) if np.isfinite(yw_orig_all[si]) else None

        x_fit_all = np.array([float((s.getPosition() if hasattr(s, "getPosition") else s.pos).x) for s in tr.catalog], dtype=float)
        y_fit_all = np.array([float((s.getPosition() if hasattr(s, "getPosition") else s.pos).y) for s in tr.catalog], dtype=float)

        ncols = len(bandnames)
        nrows = 3
        fig, axes = plt.subplots(nrows, ncols, figsize=(3 * ncols, 3 * nrows), constrained_layout=True)
        if ncols == 1:
            axes = np.array(axes).reshape(nrows, 1)

        axes[0, 0].set_ylabel("Data", fontsize=11)
        axes[1, 0].set_ylabel("Model+Sky", fontsize=11)
        axes[2, 0].set_ylabel("Residual", fontsize=11)

        for i, bn in enumerate(bandnames):
            data, (y0, y1, x0, x1) = _cut(data_by_band[i], cx, cy, size)
            model, _ = _cut(model_by_band[i], cx, cy, size)
            model = model + sky_by_band[i]
            resid = data - model

            x_fit_c = cx - float(x0)
            y_fit_c = cy - float(y0)
            x_orig_c = (x0_orig - float(x0)) if (x0_orig is not None) else None
            y_orig_c = (y0_orig - float(y0)) if (y0_orig is not None) else None

            x_fit_c_all = x_fit_all - float(x0)
            y_fit_c_all = y_fit_all - float(y0)
            ok_fit = (
                np.isfinite(x_fit_c_all)
                & np.isfinite(y_fit_c_all)
                & (x_fit_c_all >= 0)
                & (x_fit_c_all < (x1 - x0))
                & (y_fit_c_all >= 0)
                & (y_fit_c_all < (y1 - y0))
            )
            ok_fit_others = ok_fit.copy()
            if 0 <= si < len(ok_fit_others):
                ok_fit_others[si] = False

            ok_orig = None
            ok_orig_others = None
            x_orig_c_all = y_orig_c_all = None
            if x_orig_all is not None and y_orig_all is not None:
                x_orig_c_all = x_orig_all - float(x0)
                y_orig_c_all = y_orig_all - float(y0)
                ok_orig = (
                    np.isfinite(x_orig_c_all)
                    & np.isfinite(y_orig_c_all)
                    & (x_orig_c_all >= 0)
                    & (x_orig_c_all < (x1 - x0))
                    & (y_orig_c_all >= 0)
                    & (y_orig_c_all < (y1 - y0))
                )
                ok_orig_others = ok_orig.copy()
                if 0 <= si < len(ok_orig_others):
                    ok_orig_others[si] = False

            if per_band_scale:
                dvmin, dvmax = _robust_limits_percentile(data, p_lo=data_p_lo, p_hi=data_p_hi)
                mvmin, mvmax = _robust_limits_percentile(model, p_lo=model_p_lo, p_hi=model_p_hi)
                rvmin, rvmax = _robust_resid_limits_percentile(resid, p_abs=resid_p_abs)
            else:
                dvmin = dvmax = mvmin = mvmax = rvmin = rvmax = None

            ax = axes[0, i]
            im0 = ax.imshow(data, origin="lower", cmap="gray", vmin=dvmin, vmax=dvmax, interpolation="nearest")
            ax.set_title(bn, fontsize=10)
            ax.set_xticks([])
            ax.set_yticks([])
            if np.isfinite(x_fit_c) and np.isfinite(y_fit_c):
                ax.axvline(x_fit_c, color="magenta", lw=0.8, alpha=0.9)
                ax.axhline(y_fit_c, color="magenta", lw=0.8, alpha=0.9)
            if (x_orig_c is not None) and (y_orig_c is not None) and np.isfinite(x_orig_c) and np.isfinite(y_orig_c):
                ax.axvline(x_orig_c, color="lime", lw=0.8, alpha=0.9, linestyle="--")
                ax.axhline(y_orig_c, color="lime", lw=0.8, alpha=0.9, linestyle="--")
            if np.any(ok_fit_others):
                ax.scatter(
                    x_fit_c_all[ok_fit_others],
                    y_fit_c_all[ok_fit_others],
                    s=18,
                    marker="x",
                    c="deepskyblue",
                    linewidths=0.8,
                    alpha=0.8,
                )
            if ok_orig_others is not None and np.any(ok_orig_others):
                ax.scatter(
                    x_orig_c_all[ok_orig_others],
                    y_orig_c_all[ok_orig_others],
                    s=18,
                    marker="x",
                    c="orange",
                    linewidths=0.8,
                    alpha=0.8,
                )

            dmin_true, dmax_true = _true_minmax(data)
            if dmin_true is not None and dmax_true is not None:
                sm0 = ScalarMappable(norm=Normalize(vmin=dmin_true, vmax=dmax_true), cmap=im0.cmap)
                sm0.set_array([])
                cb0 = fig.colorbar(sm0, ax=ax, fraction=0.046, pad=0.02, extend="neither")
                cb0.set_label("Data")

            ax = axes[1, i]
            im1 = ax.imshow(model, origin="lower", cmap="gray", vmin=mvmin, vmax=mvmax, interpolation="nearest")
            ax.set_xticks([])
            ax.set_yticks([])
            if np.isfinite(x_fit_c) and np.isfinite(y_fit_c):
                ax.axvline(x_fit_c, color="magenta", lw=0.8, alpha=0.9)
                ax.axhline(y_fit_c, color="magenta", lw=0.8, alpha=0.9)
            if (x_orig_c is not None) and (y_orig_c is not None) and np.isfinite(x_orig_c) and np.isfinite(y_orig_c):
                ax.axvline(x_orig_c, color="lime", lw=0.8, alpha=0.9, linestyle="--")
                ax.axhline(y_orig_c, color="lime", lw=0.8, alpha=0.9, linestyle="--")
            if np.any(ok_fit_others):
                ax.scatter(
                    x_fit_c_all[ok_fit_others],
                    y_fit_c_all[ok_fit_others],
                    s=18,
                    marker="x",
                    c="deepskyblue",
                    linewidths=0.8,
                    alpha=0.8,
                )
            if ok_orig_others is not None and np.any(ok_orig_others):
                ax.scatter(
                    x_orig_c_all[ok_orig_others],
                    y_orig_c_all[ok_orig_others],
                    s=18,
                    marker="x",
                    c="orange",
                    linewidths=0.8,
                    alpha=0.8,
                )

            mmin_true, mmax_true = _true_minmax(model)
            if mmin_true is not None and mmax_true is not None:
                sm1 = ScalarMappable(norm=Normalize(vmin=mmin_true, vmax=mmax_true), cmap=im1.cmap)
                sm1.set_array([])
                cb1 = fig.colorbar(sm1, ax=ax, fraction=0.046, pad=0.02, extend="neither")
                cb1.set_label("Model+Sky")

            ax = axes[2, i]
            im2 = ax.imshow(resid, origin="lower", cmap="gray", vmin=rvmin, vmax=rvmax, interpolation="nearest")
            ax.set_xticks([])
            ax.set_yticks([])
            if np.isfinite(x_fit_c) and np.isfinite(y_fit_c):
                ax.axvline(x_fit_c, color="magenta", lw=0.8, alpha=0.9)
                ax.axhline(y_fit_c, color="magenta", lw=0.8, alpha=0.9)
            if (x_orig_c is not None) and (y_orig_c is not None) and np.isfinite(x_orig_c) and np.isfinite(y_orig_c):
                ax.axvline(x_orig_c, color="lime", lw=0.8, alpha=0.9, linestyle="--")
                ax.axhline(y_orig_c, color="lime", lw=0.8, alpha=0.9, linestyle="--")
            if np.any(ok_fit_others):
                ax.scatter(
                    x_fit_c_all[ok_fit_others],
                    y_fit_c_all[ok_fit_others],
                    s=18,
                    marker="x",
                    c="deepskyblue",
                    linewidths=0.8,
                    alpha=0.8,
                )
            if ok_orig_others is not None and np.any(ok_orig_others):
                ax.scatter(
                    x_orig_c_all[ok_orig_others],
                    y_orig_c_all[ok_orig_others],
                    s=18,
                    marker="x",
                    c="orange",
                    linewidths=0.8,
                    alpha=0.8,
                )

            rmin_true, rmax_true = _true_minmax(resid)
            if rmin_true is not None and rmax_true is not None:
                sm2 = ScalarMappable(norm=Normalize(vmin=rmin_true, vmax=rmax_true), cmap=im2.cmap)
                sm2.set_array([])
                cb2 = fig.colorbar(sm2, ax=ax, fraction=0.046, pad=0.02, extend="neither")
                cb2.set_label("Residual")

        s_id = sid if sid is not None else f"row{si}"
        s0 = f"ID={s_id} | patch={patch_tag} | idx_in_patch={si}/{len(tr.catalog)-1}"
        if stype is not None and stype != "" and stype.lower() != "nan":
            s0 += f" | TYPE={stype}"
        s_legend = "fit=center crosshair=magenta, orig=crosshair=lime(--), fit/others=x deepskyblue, orig/others=x orange"
        title = s0 + "\n" + s_legend
        fig.suptitle(title, fontsize=12)
        out_path = outdir / f"src_{si:06d}.png"
        plt.savefig(out_path, dpi=120)
        plt.close(fig)


def compute_flux_errors(tr: Tractor) -> dict[tuple[int, str], float]:
    names = tr.getParamNames()
    bandnames = [str(getattr(tim, "name", f"band{i}")) for i, tim in enumerate(tr.images)]
    band_to_i = {bn: i for i, bn in enumerate(bandnames)}

    pat = re.compile(r"^catalog\.source(\d+)\.brightness\.(.+)$")

    var = tr.optimize(variance=True, just_variance=True)
    if var is None or len(var) != len(names):
        return {}

    out: dict[tuple[int, str], float] = {}
    for i, pname in enumerate(names):
        m = pat.match(pname)
        if not m:
            continue
        sid = int(m.group(1))
        bn = m.group(2)
        if band_to_i.get(bn, None) is None:
            continue
        v = var[i]
        if v is None or (not np.isfinite(v)) or v <= 0:
            continue
        out[(sid, bn)] = float(math.sqrt(v))
    return out


def run_patch(*, payload: dict, epsf_root: Path, outdir: Path, cfg: PatchRunConfig) -> pd.DataFrame:
    meta = payload.get("meta", {}) or {}
    tag = _patch_tag_from_meta(meta)
    patch_out = outdir / tag
    _ensure_dir(patch_out)
    _ensure_dir(patch_out / "cutouts")

    log = _make_logger(patch_out, tag)
    log.info("Starting patch run: %s", tag)

    (x0_roi, x1_roi, y0_roi, y1_roi) = _roi_bbox_from_meta(meta)
    meta = dict(meta)
    meta.setdefault("roi_bbox", (int(x0_roi), int(x1_roi), int(y0_roi), int(y1_roi)))
    (patch_out / "meta.json").write_text(json.dumps(meta, indent=2, default=str))

    per_filter_patch = payload["per_filter_patch"]
    cat_patch = payload["cat_patch"]
    if not isinstance(cat_patch, pd.DataFrame):
        cat_patch = pd.DataFrame(cat_patch)

    epsf_tag = str(meta.get("epsf_tag", ""))
    if not epsf_tag:
        raise ValueError("meta.epsf_tag is required")

    psf_by_band, psf_audit = build_psf_by_band_for_patch(
        per_filter_patch=per_filter_patch,
        epsf_root=epsf_root,
        epsf_tag=epsf_tag,
        cfg=cfg,
        log=log,
    )
    tractor_images, _image_data = build_tractor_images_from_per_filter_patch(per_filter_patch=per_filter_patch, psf_by_filt=psf_by_band, use_sky_sigma=True)
    catalog = build_catalog_from_cat_patch(cat_patch, tractor_images, cfg)

    tr = Tractor(tractor_images, catalog, optimizer=ConstrainedOptimizer())
    tr.thawAllRecursive()
    # Tractor's constrained optimizer expects bounds arrays to align with thawed
    # parameter count. For GaussianMixturePSF image params this can mismatch
    # internally (params > bounds), causing IndexError in optimize().
    # Keep source params thawed, but freeze image-level GaussianMixturePSF params.
    psf_frozen_bands: list[str] = []
    for tim in tr.images:
        try:
            psf_obj = getattr(tim, "psf", None)
            if isinstance(psf_obj, GaussianMixturePSF):
                tim.freezeParam("psf")
                psf_frozen_bands.append(str(getattr(tim, "name", "")))
        except Exception as e:
            log.warning("[PSF] Failed to freeze GaussianMixturePSF params for %s: %s", str(getattr(tim, "name", "")), str(e))
    if psf_frozen_bands:
        log.warning(
            "[PSF] Frozen GaussianMixturePSF image params for optimizer stability in bands: %s",
            ",".join(psf_frozen_bands),
        )

    converged = False
    last_dlnp: float | None = None
    niters = 0
    for it in range(int(cfg.n_opt_iters)):
        dlnp, _X, _alpha = tr.optimize()
        niters = it + 1
        try:
            last_dlnp = float(dlnp)
        except Exception:
            last_dlnp = None
        log.info("opt iter %02d  dlnp=%s", it, str(dlnp))
        if dlnp < float(cfg.dlnp_stop):
            converged = True
            break

    margin = max(0, int(getattr(cfg, "flag_maxiter_margin", 0)))
    hit_max_iters = (not converged) and (niters >= max(1, int(cfg.n_opt_iters) - margin))
    if hit_max_iters:
        log.info(
            "WARNING: did not converge before max iters. niters=%d n_opt_iters=%d margin=%d last_dlnp=%s",
            int(niters),
            int(cfg.n_opt_iters),
            int(margin),
            str(last_dlnp),
        )

    meta = dict(meta)
    low_star_bands = sorted(
        [
            b
            for b, a in psf_audit.items()
            if isinstance(a.get("fallback_reason", None), str)
            and str(a.get("fallback_reason")).startswith("low-star ePSF")
        ]
    )
    fallback_bands = sorted([b for b, a in psf_audit.items() if not bool(a.get("used_epsf", False))])
    fallback_reason_by_band = {b: a.get("fallback_reason", None) for b, a in psf_audit.items() if not bool(a.get("used_epsf", False))}
    used_epsf_bands = sorted([b for b, a in psf_audit.items() if bool(a.get("used_epsf", False))])
    meta["psf_min_epsf_nstars_for_use"] = int(getattr(cfg, "min_epsf_nstars_for_use", 5))
    meta["psf_used_epsf_band_count"] = int(len(used_epsf_bands))
    meta["psf_fallback_band_count"] = int(len(fallback_bands))
    meta["psf_low_star_band_count"] = int(len(low_star_bands))
    meta["psf_fallback_bands"] = fallback_bands
    meta["psf_low_star_bands"] = low_star_bands
    meta["psf_frozen_bands_for_optimizer"] = sorted([b for b in psf_frozen_bands if b])
    meta["psf_audit"] = psf_audit
    meta["opt_niters"] = int(niters)
    meta["opt_converged"] = bool(converged)
    meta["opt_hit_max_iters"] = bool(hit_max_iters)
    meta["opt_last_dlnp"] = (float(last_dlnp) if last_dlnp is not None and np.isfinite(last_dlnp) else None)
    (patch_out / "meta.json").write_text(json.dumps(meta, indent=2, default=str))

    out = cat_patch.copy()
    out["patch_tag"] = str(tag)
    out["epsf_tag"] = str(epsf_tag)
    out["psf_min_epsf_nstars_for_use"] = int(getattr(cfg, "min_epsf_nstars_for_use", 5))
    out["psf_used_epsf_band_count"] = int(len(used_epsf_bands))
    out["psf_fallback_band_count"] = int(len(fallback_bands))
    out["psf_low_star_band_count"] = int(len(low_star_bands))
    out["psf_fallback_bands"] = ",".join(fallback_bands)
    out["psf_low_star_bands"] = ",".join(low_star_bands)
    out["psf_frozen_bands_for_optimizer"] = ",".join(sorted([b for b in psf_frozen_bands if b]))
    out["psf_fallback_reasons_json"] = json.dumps(fallback_reason_by_band, sort_keys=True)
    out["opt_niters"] = int(niters)
    out["opt_converged"] = bool(converged)
    out["opt_hit_max_iters"] = bool(hit_max_iters)
    out["opt_last_dlnp"] = float(last_dlnp) if last_dlnp is not None and np.isfinite(last_dlnp) else np.nan
    out["x_pix_patch_fit"] = np.nan
    out["y_pix_patch_fit"] = np.nan
    out["x_pix_white_fit"] = np.nan
    out["y_pix_white_fit"] = np.nan
    out["RA_fit"] = np.nan
    out["DEC_fit"] = np.nan
    out["stype_fit"] = ""
    out["sersic_n_fit"] = np.nan
    out["re_pix_fit"] = np.nan
    out["ab_fit"] = np.nan
    out["phi_deg_fit"] = np.nan
    out["ELL_fit"] = np.nan
    out["Re_fit"] = np.nan
    out["THETA_fit"] = np.nan

    bandnames = [str(getattr(tim, "name", f"band{i}")) for i, tim in enumerate(tractor_images)]
    for bn in bandnames:
        out[f"FLUX_{bn}_fit"] = np.nan
        out[f"FLUXERR_{bn}_fit"] = np.nan

    def _get_sersic_n(src, default=np.nan):
        if not hasattr(src, "sersicindex"):
            return default
        si = src.sersicindex
        if hasattr(si, "getValue"):
            return float(si.getValue())
        if hasattr(si, "value"):
            return float(si.value)
        if hasattr(si, "n"):
            return float(si.n)
        return default

    def _get_shape_params(src):
        if not hasattr(src, "shape"):
            return (np.nan, np.nan, np.nan)
        sh = src.shape
        re_ = float(getattr(sh, "re", np.nan))
        ab_ = float(getattr(sh, "ab", np.nan))
        if hasattr(sh, "phi"):
            phi_deg = float(getattr(sh, "phi", np.nan))
        elif hasattr(sh, "theta"):
            try:
                phi_deg = float(-np.degrees(sh.theta))
            except Exception:
                phi_deg = np.nan
        else:
            phi_deg = np.nan
        if np.isfinite(phi_deg):
            phi_deg = float(phi_deg % 180.0)
        return (re_, ab_, phi_deg)

    for i, src in enumerate(tr.catalog):
        pos = src.getPosition() if hasattr(src, "getPosition") else src.pos
        bx = src.getBrightness() if hasattr(src, "getBrightness") else src.brightness
        out.at[i, "x_pix_patch_fit"] = float(pos.x)
        out.at[i, "y_pix_patch_fit"] = float(pos.y)
        out.at[i, "x_pix_white_fit"] = float(pos.x) + float(x0_roi)
        out.at[i, "y_pix_white_fit"] = float(pos.y) + float(y0_roi)

        if isinstance(src, SersicGalaxy):
            out.at[i, "stype_fit"] = "sersic"
            out.at[i, "sersic_n_fit"] = _get_sersic_n(src)
            re_, ab_, phi_ = _get_shape_params(src)
            out.at[i, "re_pix_fit"] = re_
            out.at[i, "ab_fit"] = ab_
            out.at[i, "phi_deg_fit"] = phi_
            if np.isfinite(re_):
                out.at[i, "Re_fit"] = float(re_)
            if np.isfinite(ab_):
                out.at[i, "ELL_fit"] = float(1.0 - ab_)
            if np.isfinite(phi_):
                out.at[i, "THETA_fit"] = float(phi_ - 90.0)
        elif isinstance(src, DevGalaxy):
            out.at[i, "stype_fit"] = "dev"
            re_, ab_, phi_ = _get_shape_params(src)
            out.at[i, "re_pix_fit"] = re_
            out.at[i, "ab_fit"] = ab_
            out.at[i, "phi_deg_fit"] = phi_
            if np.isfinite(re_):
                out.at[i, "Re_fit"] = float(re_)
            if np.isfinite(ab_):
                out.at[i, "ELL_fit"] = float(1.0 - ab_)
            if np.isfinite(phi_):
                out.at[i, "THETA_fit"] = float(phi_ - 90.0)
        elif isinstance(src, ExpGalaxy):
            out.at[i, "stype_fit"] = "exp"
            re_, ab_, phi_ = _get_shape_params(src)
            out.at[i, "re_pix_fit"] = re_
            out.at[i, "ab_fit"] = ab_
            out.at[i, "phi_deg_fit"] = phi_
            if np.isfinite(re_):
                out.at[i, "Re_fit"] = float(re_)
            if np.isfinite(ab_):
                out.at[i, "ELL_fit"] = float(1.0 - ab_)
            if np.isfinite(phi_):
                out.at[i, "THETA_fit"] = float(phi_ - 90.0)
        else:
            out.at[i, "stype_fit"] = "star"

        for bn in bandnames:
            try:
                out.at[i, f"FLUX_{bn}_fit"] = float(bx.getFlux(bn))
            except Exception:
                out.at[i, f"FLUX_{bn}_fit"] = np.nan

    try:
        ferr = compute_flux_errors(tr)
        for (sid, bn), fe in ferr.items():
            if 0 <= sid < len(out):
                out.at[sid, f"FLUXERR_{bn}_fit"] = float(fe)
        log.info("Computed flux errors for %d (source,band) pairs", len(ferr))
    except Exception as e:
        log.info("WARNING: flux error computation failed: %s", str(e))

    out_csv = patch_out / f"{tag}_cat_fit.csv"
    out.to_csv(out_csv, index=False)
    log.info("Wrote fitted catalog: %s", str(out_csv))

    if cfg.save_cutouts:
        save_source_montages(
            tr=tr,
            cat_patch=cat_patch,
            meta=meta,
            outdir=patch_out / "cutouts",
            size=int(cfg.cutout_size_pix),
            per_band_scale=True,
            max_sources=cfg.cutout_max_sources,
            start=int(cfg.cutout_start),
            data_p_lo=float(cfg.cutout_data_p_lo),
            data_p_hi=float(cfg.cutout_data_p_hi),
            model_p_lo=float(cfg.cutout_model_p_lo),
            model_p_hi=float(cfg.cutout_model_p_hi),
            resid_p_abs=float(cfg.cutout_resid_p_abs),
            log=log,
        )

    if getattr(cfg, "save_patch_overview", False):
        try:
            save_patch_overview(
                tr=tr,
                cat_patch=cat_patch,
                meta=meta,
                outpath=patch_out / "patch_overview.png",
                data_p_lo=float(cfg.cutout_data_p_lo),
                data_p_hi=float(cfg.cutout_data_p_hi),
                model_p_lo=float(cfg.cutout_model_p_lo),
                model_p_hi=float(cfg.cutout_model_p_hi),
                resid_p_abs=float(cfg.cutout_resid_p_abs),
                log=log,
            )
        except Exception as e:
            log.info("WARNING: patch overview failed: %s", str(e))

    log.info("Done patch run: %s", tag)
    return out


def main(argv: Optional[list[str]] = None) -> int:
    ap = argparse.ArgumentParser()
    ap.add_argument("--inputs", required=True, help="Path to patch input .pkl.gz")
    ap.add_argument("--epsf-root", required=True, help="ePSF root directory")
    ap.add_argument("--outdir", required=True, help="Output base directory")
    ap.add_argument("--no-cutouts", action="store_true", help="Disable montage saving")
    ap.add_argument("--no-patch-overview", action="store_true", help="Disable per-patch overview plot")
    ap.add_argument("--cutout-size", type=int, default=100)
    ap.add_argument("--cutout-max-sources", type=int, default=None)
    ap.add_argument("--cutout-start", type=int, default=0)
    ap.add_argument("--cutout-data-p-lo", type=float, default=1)
    ap.add_argument("--cutout-data-p-hi", type=float, default=99)
    ap.add_argument("--cutout-model-p-lo", type=float, default=1)
    ap.add_argument("--cutout-model-p-hi", type=float, default=99)
    ap.add_argument("--cutout-resid-p-abs", type=float, default=99)
    ap.add_argument("--gal-model", default="dev", choices=["dev", "exp", "sersic", "star"])
    ap.add_argument("--n-opt-iters", type=int, default=200)
    ap.add_argument("--dlnp-stop", type=float, default=1e-6)
    ap.add_argument("--flag-maxiter-margin", type=int, default=5)
    ap.add_argument("--r-ap", type=float, default=5.0)
    ap.add_argument("--eps-flux", type=float, default=1e-4)
    ap.add_argument("--psf-model", default="hybrid", choices=["hybrid", "pixelized", "gaussian_mixture_from_stamp"])
    ap.add_argument("--psf-fallback-model", default="gaussian_mixture", choices=["gaussian_mixture", "ncircular_gaussian", "moffat"])
    ap.add_argument("--min-epsf-nstars-for-use", type=int, default=5)
    ap.add_argument("--ncircular-gaussian-n", type=int, default=3, choices=[1, 2, 3])
    ap.add_argument("--hybrid-psf-n", type=int, default=4)
    ap.add_argument("--gaussian-mixture-n", type=int, default=4)
    ap.add_argument("--sersic-n-init", type=float, default=3.0)
    ap.add_argument("--re-fallback-pix", type=float, default=3.0)
    ap.add_argument("--fallback-default-fwhm-pix", type=float, default=4.0)
    ap.add_argument("--fallback-stamp-radius-pix", type=float, default=25.0)
    ap.add_argument("--moffat-beta", type=float, default=3.5)
    ap.add_argument("--moffat-radius-pix", type=float, default=25.0)
    args = ap.parse_args(argv)

    payload = load_patch_inputs(args.inputs)
    cfg = PatchRunConfig(
        save_cutouts=not args.no_cutouts,
        save_patch_overview=not args.no_patch_overview,
        cutout_size_pix=int(args.cutout_size),
        cutout_max_sources=args.cutout_max_sources,
        cutout_start=int(args.cutout_start),
        cutout_data_p_lo=float(args.cutout_data_p_lo),
        cutout_data_p_hi=float(args.cutout_data_p_hi),
        cutout_model_p_lo=float(args.cutout_model_p_lo),
        cutout_model_p_hi=float(args.cutout_model_p_hi),
        cutout_resid_p_abs=float(args.cutout_resid_p_abs),
        gal_model=str(args.gal_model),
        n_opt_iters=int(args.n_opt_iters),
        dlnp_stop=float(args.dlnp_stop),
        flag_maxiter_margin=int(args.flag_maxiter_margin),
        r_ap=float(args.r_ap),
        eps_flux=float(args.eps_flux),
        psf_model=str(args.psf_model),
        psf_fallback_model=str(args.psf_fallback_model),
        min_epsf_nstars_for_use=int(args.min_epsf_nstars_for_use),
        ncircular_gaussian_n=int(args.ncircular_gaussian_n),
        hybrid_psf_n=int(args.hybrid_psf_n),
        gaussian_mixture_n=int(args.gaussian_mixture_n),
        sersic_n_init=float(args.sersic_n_init),
        re_fallback_pix=float(args.re_fallback_pix),
        fallback_default_fwhm_pix=float(args.fallback_default_fwhm_pix),
        fallback_stamp_radius_pix=float(args.fallback_stamp_radius_pix),
        moffat_beta=float(args.moffat_beta),
        moffat_radius_pix=float(args.moffat_radius_pix),
    )

    run_patch(payload=payload, epsf_root=Path(args.epsf_root), outdir=Path(args.outdir), cfg=cfg)
    return 0


if __name__ == "__main__":
    raise SystemExit(main())
